# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import os
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from lewis.scripts.control import control_simulation  # noqa: E402

if __name__ == "__main__":
    control_simulation()



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from lewis.devices.julabo.devices.device import SimulatedJulabo

from .utils import assertRaisesNothing


class TestSimulatedJulabo(unittest.TestCase):
    def setUp(self):
        self.julabo_device = SimulatedJulabo()
        self.julabo_device.process()  # Initialise

    def go_forward_one_minute(self):
        """Moves the device forward in time by 60 seconds."""
        self.julabo_device.process(60)

    def check_setting_values_works(self, par_name, value):
        """Helper function for checking that setting a simple value works.

        :param param_name: The name of the parameter to change.
        :param value: The new value.
        """
        setter = getattr(self.julabo_device, "set_%s" % par_name)
        setter(value)
        self.go_forward_one_minute()
        self.assertEqual(value, getattr(self.julabo_device, par_name))

    def test_default_construction(self):
        assertRaisesNothing(self, SimulatedJulabo)

    def test_changing_setpoint_changes_setpoint(self):
        old_sp = self.julabo_device.set_point_temperature

        self.julabo_device.set_set_point(old_sp + 10.0)

        self.assertEqual(old_sp + 10.0, self.julabo_device.set_point_temperature)

    def test_on_initialisation_circulating_is_off(self):
        self.assertEqual(0, self.julabo_device.is_circulating)

    def test_on_stating_circulating_it_starts_circulating(self):
        self.julabo_device.set_circulating(1)
        self.assertEqual(1, self.julabo_device.is_circulating)

    def test_changing_setpoint_does_not_change_temperature_if_not_circulating(self):
        old_sp = self.julabo_device.set_point_temperature
        old_t = self.julabo_device.temperature

        self.julabo_device.set_set_point(old_sp + 10.0)
        self.go_forward_one_minute()

        self.assertEqual(old_t, self.julabo_device.temperature)

    def test_changing_setpoint_does_change_temperature_if_circulating(self):
        self.julabo_device = SimulatedJulabo()
        self.julabo_device.process()  # Initialise
        old_sp = self.julabo_device.set_point_temperature
        old_t = self.julabo_device.temperature

        self.julabo_device.set_set_point(old_sp + 10.0)
        self.julabo_device.set_circulating(1)
        self.go_forward_one_minute()

        self.assertNotEqual(old_t, self.julabo_device.temperature)

    def test_setting_external_p_sets_p(self):
        self.check_setting_values_works("external_p", 10)

    def test_setting_external_i_sets_i(self):
        self.check_setting_values_works("external_i", 10)

    def test_setting_external_d_sets_d(self):
        self.check_setting_values_works("external_d", 10)

    def test_setting_internal_p_sets_p(self):
        self.check_setting_values_works("internal_p", 10)

    def test_setting_internal_i_sets_i(self):
        self.check_setting_values_works("internal_i", 10)

    def test_setting_internal_d_sets_d(self):
        self.check_setting_values_works("internal_d", 10)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.core import approaches
from lewis.core.statemachine import State


class DefaultNotCirculatingState(State):
    pass


class DefaultCirculatingState(State):
    def in_state(self, dt) -> None:
        # Approach target temperature at a set rate
        self._context.temperature = approaches.linear(
            self._context.temperature,
            self._context.set_point_temperature,
            self._context.heating_power / 60.0,
            dt,
        )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
Defines functions that model typical behavior, such as a value approaching a target linearly at
a certain rate.
"""


def linear(current: float, target: float, rate: float, dt: float):
    """
    This function returns the new value after moving towards
    target at the given speed constantly for the time dt.

    If for example the current position is 10 and the target is -20,
    the returned value will be less than 10 if rate and dt are greater
    than 0:

    .. sourcecode:: Python

        new_pos = linear(10, -20, 10, 0.1)  # new_pos = 9

    The function makes sure that the returned value never overshoots:

    .. sourcecode:: Python

        new_pos = linear(10, -20, 10, 100)  # new_pos = -20

    :param current: The current value of the variable to be changed.
    :param target: The target value to approach.
    :param rate: The rate at which the parameter should move towards target.
    :param dt: The time for which to calculate the change.
    :return: The new variable value.
    """
    sign = (target > current) - (target < current)

    if not sign:
        return current

    new_value = current + sign * rate * dt

    if sign * new_value > sign * target:
        return target

    return new_value



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module provides client code for objects exposed via JSON-RPC over ZMQ.

.. seealso::

    The server-part for these client classes is defined
    in the module :mod:`~lewis.core.control_server`.
"""

import types
import uuid

import zmq

# This does not import .exceptions, because absolute_import from the __future__ module
try:
    import exceptions
except ImportError:
    import builtins as exceptions


class RemoteException(Exception):
    """
    This exception type replaces exceptions that are raised on the server,
    but unknown (i.e. not in the exceptions-module) on the client side.
    To retain as much information as possible, the exception type on the server and
    the message are stored.

    :param exception_type: Type of the exception on the server side.
    :param message: Exception message on the server side.
    """

    def __init__(self, exception_type, message) -> None:
        super(RemoteException, self).__init__(
            "Exception on server side of type '{}': '{}'".format(exception_type, message)
        )

        self.server_side_type = exception_type
        self.server_side_message = message


class ProtocolException(Exception):
    """
    An exception type for exceptions related to the transport protocol, i.e.
    malformed requests etc.
    """


class ControlClient:
    """
    This class provides an interface to a ControlServer instance on
    the server side. Proxies to exposed objects can be obtained either
    directly via get_object or, in case the server exposes a collection
    of objects at the top level, a dictionary of named objects can be
    obtained via get_object_collection.

    If a timeout is supplied, all underlying network operations time out
    after the specified time (in milliseconds), for no timeout specify ``None``.

    :param host: Host the control server is running on.
    :param port: Port on which the control server is listening.
    :param timeout: Timeout in milliseconds for ZMQ operations.
    """

    def __init__(self, host="127.0.0.1", port="10000", timeout=3000) -> None:
        self.timeout = timeout if timeout is not None else -1

        self._socket = self._get_zmq_req_socket()

        self._connection_string = "tcp://{0}:{1}".format(host, port)
        self._socket.connect(self._connection_string)

    def _get_zmq_req_socket(self):
        context = zmq.Context()
        context.setsockopt(zmq.REQ_CORRELATE, 1)
        context.setsockopt(zmq.REQ_RELAXED, 1)
        context.setsockopt(zmq.SNDTIMEO, self.timeout)
        context.setsockopt(zmq.RCVTIMEO, self.timeout)
        context.setsockopt(zmq.LINGER, 0)
        return context.socket(zmq.REQ)

    def json_rpc(self, method, *args):
        """
        This method takes a ZMQ REQ-socket and submits a JSON-object containing
        the RPC (JSON-RPC 2.0 format) to the supplied method with the supplied arguments.
        Then it waits for a reply from the server and blocks until it has received
        a JSON-response. The method returns the response and the id it used to tag
        the original request, which is a random UUID (uuid.uuid4).

        :param method: Method to call on remote.
        :param args: Arguments to method call.
        :return: JSON result and request id.
        """
        request_id = str(uuid.uuid4())

        try:
            self._socket.send_json(
                {"method": method, "params": args, "jsonrpc": "2.0", "id": request_id}
            )

            return self._socket.recv_json(), request_id
        except zmq.error.Again:
            raise ProtocolException(
                "The ZMQ connection to {} timed out after {:.2f}s.".format(
                    self._connection_string, self.timeout / 1000
                )
            )

    def get_object(self, object_name=""):
        api, request_id = self.json_rpc(object_name + ":api")

        if "result" not in api or api["id"] != request_id:
            raise ProtocolException("Failed to retrieve API of remote object.")

        object_type = type(str(api["result"]["class"]), (ObjectProxy,), {})
        methods = api["result"]["methods"]

        glue = "." if object_name else ""
        return object_type(self, methods, object_name + glue)

    def get_object_collection(self, object_name=""):
        """
        If the remote end exposes a collection of objects under the supplied object name (empty
        for top level), this method returns a dictionary of these objects stored under their
        names on the server.

        This function performs n + 1 calls to the server, where n is the number of objects.

        :param object_name: Object name on the server. This is required if the object collection
                            is not the top level object.
        """

        object_names = self.get_object(object_name).get_objects()

        return {obj: self.get_object(obj) for obj in object_names}


class ObjectProxy:
    """
    This class serves as a base class for dynamically created classes on the
    client side that represent server-side objects. Upon initialization,
    this class takes the supplied methods and installs appropriate proxy methods
    or properties into the object and class respectively. Because of that
    class manipulation, this class must never be used directly.
    Instead, it should be used as a base-class for dynamically created types
    that mirror types on the server, like this:

    .. sourcecode:: Python

        proxy = type("SomeClassName", (ObjectProxy,), {})(connection, methods, prefix)

    There is however, the class ControlClient, which automates all that
    and provides objects that are ready to use.

    Exceptions on the server are propagated to the client. If the exception is not part
    of the exceptions-module (builtins for Python 3), a RemoteException is raised instead
    which contains information about the server side exception.

    All RPC method names are prefixed with the supplied prefix, which is usually the
    object name on the server plus a dot.

    :param connection: ControlClient-object for remote calls.
    :param members: List of strings to generate methods and properties.
    :param prefix: Usually object name on the server plus dot.
    """

    def __init__(self, connection, members, prefix="") -> None:
        self._properties = set()

        self._connection = connection
        self._prefix = prefix
        self._add_member_proxies(members)

    def _make_request(self, method, *args):
        """
        This method performs a JSON-RPC request via the object's ZMQ socket. If successful,
        the result is returned, otherwise exceptions are raised. Server side exceptions are
        raised using the same type as on the server if they are part of the exceptions-module.
        Otherwise, a RemoteException is raised.

        :param method: Method of the object to call on the remote.
        :param args: Positional arguments to the method call.
        :return: Result of the remote call if successful.
        """
        response, request_id = self._connection.json_rpc(self._prefix + method, *args)

        if "id" not in response:
            raise ProtocolException("JSON-RPC response does not contain ID field.")

        if response["id"] != request_id:
            raise ProtocolException(
                "ID of JSON-RPC request ({}) did not match response ({}).".format(
                    request_id, response["id"]
                )
            )

        if "result" in response:
            return response["result"]

        if "error" in response:
            if "data" in response["error"]:
                exception_type = response["error"]["data"]["type"]
                exception_message = response["error"]["data"]["message"]

                if not hasattr(exceptions, exception_type):
                    raise RemoteException(exception_type, exception_message)
                else:
                    exception = getattr(exceptions, exception_type)
                    raise exception(exception_message)
            else:
                raise ProtocolException(response["error"]["message"])

    def _add_member_proxies(self, members) -> None:
        for member in [str(m) for m in members]:
            if ":set" in member or ":get" in member:
                self._properties.add(member.split(":")[-2].split(".")[-1])
            else:
                setattr(self, member, self._create_method_proxy(member))

        for prop in self._properties:
            setattr(
                type(self),
                prop,
                property(self._create_getter_proxy(prop), self._create_setter_proxy(prop)),
            )

    def _create_getter_proxy(self, property_name):
        def getter(obj):
            return obj._make_request(property_name + ":get")

        return getter

    def _create_setter_proxy(self, property_name):
        def setter(obj, value):
            return obj._make_request(property_name + ":set", value)

        return setter

    def _create_method_proxy(self, method_name):
        def method_wrapper(obj, *args):
            return obj._make_request(method_name, *args)

        return types.MethodType(method_wrapper, self)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
A :class:`~Simulation` combines a :mod:`Device <lewis.devices>` and its interface (derived from
an :mod:`Adapter <lewis.adapters>`).
"""

from datetime import datetime
from threading import Thread
from time import sleep

from lewis.core.adapters import AdapterCollection
from lewis.core.control_server import ControlServer, ExposedObject
from lewis.core.devices import DeviceRegistry
from lewis.core.logging import has_log
from lewis.core.utils import seconds_since


@has_log
class Simulation:
    """
    The Simulation class controls certain aspects of a device simulation,
    the most important one being time.

    Once :meth:`start` is called, the process-method of the device
    is called in regular intervals. The time between these calls is
    influenced by the cycle_delay property. Because of the way some
    network protocols work, the actual processing time can be
    longer or shorter, so cycle_delay should be seen as a guideline
    rather than a guaranteed parameter.

    In the simplest case, the actual time-delta between two cycles
    is passed to the simulated device so that it can update its internal
    state according to the elapsed time. It is however possible to set
    a simulation speed, which serves as a multiplier for this time.
    If the speed is set to 2 and 0.1 seconds pass between two cycles,
    the simulation is asked to simulate 0.2 seconds, and so on. Speed 0
    effectively stops all time dependent calculations in the
    simulated device.

    Another possibility to pause the simulation is the pause-method. After
    calling it, all processing in the device is suspended, while the communication
    adapters continue to work. This can be used to simulate that a device is "hanging".
    The simulation can be continued using the resume-method.

    A number of status properties provide information about the simulation.
    The total uptime (in actually elapsed time) can be obtained through the
    uptime-property, whereas the runtime-property contains the simulated time.
    The cycles-property indicates the total number of simulation cycles, which
    does not increase when the simulation is paused.

    Finally, the simulation can be stopped entirely with the stop-method.

    All functionality except for the start-method can be made available to remote
    computers via a :class:`ControlServer`-instance. The way to expose device and simulation
    is to pass a 'host:port'-string as the control_server argument,
    which will construct the control server. Simulation will try to start the
    control server using the start_server method.

    :param device: The simulated device.
    :param adapters: Adapters which expose the simulated device.
    :param device_builder: :class:`~lewis.core.devices.DeviceBuilder` instance to enable setup-
                           switching at runtime.
    :param control_server: 'host:port'-string to construct control server or None.
    """

    def __init__(self, device, adapters=(), device_builder=None, control_server=None) -> None:
        super(Simulation, self).__init__()

        self._device_builder = device_builder

        self._device = device
        self._adapters = AdapterCollection(*adapters)

        self._speed = 1.0  # Multiplier for delta t
        self._cycle_delay = 0.1  # Target time between cycles

        self._start_time = None  # Real time when the simulation started
        self._cycles = 0  # Number of cycles processed
        self._runtime = 0.0  # Total simulation time processed

        self._running = False
        self._started = False
        self._stop_commanded = False

        # Constructing the control server must be deferred until the end,
        # because the construction is not complete at this point
        self._control_server = None  # Just initialize to None and use property setter afterwards
        self._control_server_thread = None
        self.control_server = control_server

        self.log.debug(
            "Created simulation. Device type: %s, Protocol(s): %s, Possible setups for "
            "switching: %s, Control server: %s",
            device.__class__.__name__,
            ", ".join(self._adapters.protocols),
            ", ".join(device_builder.setups.keys()) if device_builder else None,
            control_server,
        )

    def _create_control_server(self, control_server):
        if control_server is None:
            return None

        return ControlServer(
            {
                "device": ExposedObject(
                    self._device,
                    exclude_inherited=True,
                    lock=self._adapters.device_lock,
                ),
                "simulation": ExposedObject(
                    self,
                    exclude=("start", "control_server", "log"),
                    exclude_inherited=True,
                ),
                "interface": ExposedObject(
                    self._adapters,
                    exclude=(
                        "device_lock",
                        "add_adapter",
                        "remove_adapter",
                        "handle",
                        "log",
                    ),
                    exclude_inherited=True,
                ),
            },
            control_server,
        )

    @property
    def setups(self):
        """
        A list of setups that are available. Use :meth:`switch_setup` to
        change the setup.
        """
        return list(self._device_builder.setups.keys()) if self._device_builder is not None else []

    def switch_setup(self, new_setup) -> None:
        """
        This method switches the setup, which means that it replaces the currently
        simulated device with a new device, as defined by the setup.

        If any error occurs during setup switching it is logged and re-raised.

        :param new_setup: Name of the new setup to load.
        """
        try:
            self._device = self._device_builder.create_device(new_setup)
            self._adapters.set_device(self._device)
            self.log.info("Switched setup to '%s'", new_setup)
        except Exception as e:
            self.log.error(
                "Caught an error while trying to switch setups. Setup not switched, "
                "simulation continues: %s",
                e,
            )
            raise

    def start(self) -> None:
        """
        Starts the simulation.
        """
        self.log.info("Starting simulation")

        self._running = True
        self._started = True
        self._stop_commanded = False

        self._start_control_server()

        self._adapters.connect()

        self._start_time = datetime.now()

        delta = 0.0

        while not self._stop_commanded:
            delta = self._process_cycle(delta)

        self._running = False
        self._started = False

        self.log.info("Simulation has ended.")

    def _start_control_server(self) -> None:
        if self._control_server is not None and self._control_server_thread is None:

            def control_server_loop() -> None:
                self._control_server.start_server()

                while not self._stop_commanded:
                    self._control_server.process(blocking=True)

                self.log.info("Stopped processing control server commands, ending thread.")

            self._control_server_thread = Thread(target=control_server_loop)
            self._control_server_thread.start()

    def _stop_control_server(self) -> None:
        if self._control_server_thread is not None:
            self._control_server_thread.join(timeout=1.0)
            self._control_server_thread = None

    def _process_cycle(self, delta):
        """
        Processes one cycle, which consists of one simulation cycle and processing
        of control server commands. The method measures how long all this takes
        and returns the elapsed time in seconds.

        :param delta: Elapsed time in last cycle, passed to simulation.
        :return: Elapsed time in this cycle.
        """
        start = datetime.now()

        self._process_simulation_cycle(delta)

        delta = seconds_since(start)

        return delta

    def _process_simulation_cycle(self, delta) -> None:
        """
        If the simulation is not paused, the device's process-method is
        called with the supplied delta, multiplied by the simulation speed.

        If the simulation is paused, the process sleeps for the duration
        of one cycle_delay.

        :param delta: Time delta passed to simulation.
        """
        self.log.debug("Cycle, dt=%s", delta)

        sleep(self._cycle_delay)

        if self._running:
            delta_simulation = delta * self._speed

            with self._adapters.device_lock:
                self._device.process(delta_simulation)

            self._cycles += 1
            self._runtime += delta_simulation

    @property
    def cycle_delay(self):
        """
        Desired time between simulation cycles, this can not be negative.
        Use 0 for highest possible processing rate.
        """
        return self._cycle_delay

    @cycle_delay.setter
    def cycle_delay(self, delay) -> None:
        if delay < 0.0:
            raise ValueError("Cycle delay can not be negative.")

        self._cycle_delay = delay

        self.log.info("Changed cycle delay to %s", self._cycle_delay)

    @property
    def cycles(self):
        """
        Simulation cycles processed since start has been called.
        """
        return self._cycles

    @property
    def uptime(self):
        """
        Elapsed time in seconds since the simulation has been started.
        """
        if not self._started:
            return 0.0
        return seconds_since(self._start_time)

    @property
    def speed(self):
        """
        Simulation speed. Actual elapsed time is multiplied with this property
        to determine simulated time. Values greater than 1 increase the simulation
        speed, values between 1 and 0 decrease it. A speed of 0 effectively pauses
        the simulation.
        """
        return self._speed

    @speed.setter
    def speed(self, new_speed) -> None:
        if new_speed < 0:
            raise ValueError("Speed can not be negative.")

        self._speed = new_speed

        self.log.info("Changed speed to %s", self._speed)

    @property
    def runtime(self):
        """
        The accumulated simulation time. Whenever speed is different from 1, this
        progresses at a different rate than uptime.
        """
        return self._runtime

    def set_device_parameters(self, parameters) -> None:
        """
        Set multiple parameters of the simulated device "simultaneously". The passed
        parameter is assumed to be device parameter/value dict.
        The method only allows to set existing attributes. If there are invalid
        attribute names, the attributes are not updated, instead a RuntimeError
        is raised. The same happens if any of the parameters are methods, which
        can not be updated with this mechanisms.

        :param parameters: Dict of device attribute/values to update the device.
        """
        invalid_parameters = set(parameters.keys()) - set(
            x for x in dir(self._device) if not callable(getattr(self._device, x))
        )
        if invalid_parameters:
            raise RuntimeError(
                "The following parameters do not exist in the device or are methods: {}."
                "Parameters not updated.".format(invalid_parameters)
            )

        with self._adapters.device_lock:
            for name, value in parameters.items():
                setattr(self._device, name, value)

        self.log.debug("Updated device parameters: %s", parameters)

    def pause(self) -> None:
        """
        Pause the simulation. Can only be called after start has been called.
        """
        if not self._running:
            raise RuntimeError("Can only pause a running simulation.")

        self.log.info("Pausing simulation")

        self._running = False

    def resume(self) -> None:
        """
        Resume a paused simulation. Can only be called after start
        and pause have been called.
        """
        if not self._started or self._running:
            raise RuntimeError("Can only resume a paused simulation.")

        self.log.info("Resuming simulation")

        self._running = True

    def stop(self) -> None:
        """
        Stops the simulation entirely.
        """
        if self.is_started:
            self.log.warning("Stopping simulation")

            self._stop_commanded = True

            self._stop_control_server()
            self._adapters.disconnect()

    @property
    def is_started(self):
        """
        This property is true if the simulation has been started.
        """
        return self._started

    @property
    def is_paused(self):
        """
        True if the simulation is paused (implies that the simulation has been started).
        """
        return self._started and not self._running

    @property
    def control_server(self):
        """
        ControlServer-instance that exposes the object to remote machines. Can only
        be set before start has been called or on a running simulation if no
        control server was previously present. If the server is not running, it will be started
        after it has been set.
        """
        return self._control_server

    @control_server.setter
    def control_server(self, control_server) -> None:
        if self.is_started and self._control_server:
            raise RuntimeError("Can not replace control server while simulation is running.")

        self._control_server = self._create_control_server(control_server)

        if self.is_started and self._control_server is not None:
            self._control_server.start_server()


class SimulationFactory:
    """
    This class is used to create :class:`Simulation`-objects according to a certain
    set of parameters, such as device, setup and protocol. To create a simulation, it needs to
    know where devices are stored:

    .. sourcecode:: Python

        factory = SimulationFactory("lewis.devices")

    The actual creation happens via the :meth:`create`-method:

    .. sourcecode:: Python

        simulation = factory.create("device_name", protocol="protocol")

    The simulation can then be started and stopped as desired.

    .. warning:: This class is meant for internal use at the moment and may change frequently.
    """

    def __init__(self, devices_package) -> None:
        self._reg = DeviceRegistry(devices_package)

    @property
    def devices(self):
        """Names of available devices."""
        return self._reg.devices

    def get_protocols(self, device):
        """Returns a list of available protocols for the specified device."""
        return self._reg.device_builder(device).protocols

    def create(self, device, setup=None, protocols=None, control_server=None):
        """
        Creates a :class:`Simulation` according to the supplied parameters.

        :param device: Name of device.
        :param setup: Name of the setup for device creation.
        :param protocols: Dictionary where each key is assigned a dictionary with options for the
                          corresponding :class:`~lewis.core.adapters.Adapter`. For available
                          protocols, see :meth:`get_protocols`.
        :param control_server: String to construct a control server (host:port).
        :return: Simulation object according to input parameters.
        """

        device_builder = self._reg.device_builder(device)
        device = device_builder.create_device(setup)

        adapters = []

        if protocols is not None:
            for protocol, options in protocols.items():
                interface = device_builder.create_interface(protocol)
                interface.device = device

                adapter = interface.adapter(options=options or {})
                adapter.interface = interface

                adapters.append(adapter)

        return Simulation(
            device=device,
            adapters=adapters,
            device_builder=device_builder,
            control_server=control_server,
        )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************


class Bearings:
    def __init__(self) -> None:
        pass

    def disengage(self) -> None:
        pass

    def engage(self) -> None:
        pass


class MagneticBearings(Bearings):
    def __init__(self) -> None:
        super(MagneticBearings, self).__init__()


class MechanicalBearings(Bearings):
    def __init__(self) -> None:
        super(MechanicalBearings, self).__init__()



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import os
import shutil
import sys
import tempfile
import unittest


def assertRaisesNothing(testobj, func, *args, **kwargs):
    """
    unittest does not have an assertRaisesNothing. This function adopted from
    the Mantid testhelpers module provides that functionality.

    :param testobj: A unittest object
    :param func: A callable object
    :param *args: Positional arguments passed to the callable as they are
    :param **kwargs: Keyword arguments, passed on as they are
    """
    try:
        return func(*args, **kwargs)
    except Exception as exc:
        testobj.fail(
            "Assertion error. An exception was caught where none "
            "was expected in %s. Message: %s" % (func.__name__, str(exc))
        )


class TestWithPackageStructure(unittest.TestCase):
    """
    This is an intermediate class that creates a package structure in the
    system's temporary file directory. The structure is as follows:

        tmp_dir (random name)
         |
         +- some_dir
         |   |
         |   +- __init__.py
         +- .invalid
         +- _invalid
         +- some_file.py
         +- some_other_file.pyc
         +- _some_invalid_file.py
         +- __init__.py

    All files are empty and the entire structure is deleted in the tearDown.
    """

    @classmethod
    def setUpClass(cls):
        cls._tmp_dir = tempfile.mkdtemp()
        cls._tmp_package = tempfile.mkdtemp(dir=cls._tmp_dir)
        cls._tmp_package_name = os.path.basename(cls._tmp_package)

        cls._files = {
            k: os.path.join(cls._tmp_package, v)
            for k, v in dict(
                valid="some_file.py",
                invalid_ext="some_other_file.pyc",
                invalid_name="_some_invalid_file.py",
                failing_module="failing_module.py",
            ).items()
        }

        for abs_file_name in cls._files.values():
            with open(abs_file_name, mode="w"):
                pass

        with open(cls._files["failing_module"], mode="w") as fh:
            fh.write("raise ImportError()\n")

        cls._dirs = {
            k: os.path.join(cls._tmp_package, v)
            for k, v in dict(
                valid="some_dir", invalid_underscore="_invalid", invalid_dot=".invalid"
            ).items()
        }

        for abs_dir_name in cls._dirs.values():
            os.mkdir(abs_dir_name)

        with open(os.path.join(cls._tmp_package, "__init__.py"), "w"):
            pass

        with open(os.path.join(cls._tmp_package, "some_dir", "__init__.py"), "w"):
            pass

        cls._expected_modules = ["some_dir", "some_file"]

        sys.path.insert(0, cls._tmp_dir)

    @classmethod
    def tearDownClass(cls):
        sys.path.pop(sys.path.index(cls._tmp_dir))
        shutil.rmtree(cls._tmp_dir)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import argparse
import os
import sys

import yaml

from lewis import __version__
from lewis.core.exceptions import LewisException
from lewis.core.logging import default_log_format, logging
from lewis.core.simulation import SimulationFactory
from lewis.scripts import get_usage_text

parser = argparse.ArgumentParser(
    description="This script starts a simulated device that is exposed via the specified "
    "communication protocol. Complete documentation of Lewis is available in "
    "the online documentation on GitHub https://github.com/ess-dmsc/lewis/",
    add_help=False,
    prog="lewis",
)

positional_args = parser.add_argument_group("Positional arguments")

positional_args.add_argument(
    "device",
    nargs="?",
    help="Name of the device to simulate, omitting this argument prints out a list "
    "of available devices.",
)

device_args = parser.add_argument_group(
    "Device related parameters",
    "Parameters that influence the selected device, such as setup or protocol.",
)

device_args.add_argument(
    "-s",
    "--setup",
    default=None,
    help="Name of the setup to load. If not provided, the default setup is selected. If there"
    "is no default, a list of setups is printed.",
)

interface_args = device_args.add_mutually_exclusive_group()
interface_args.add_argument(
    "-n",
    "--no-interface",
    default=False,
    action="store_true",
    help="If supplied, the device simulation will not have any communication interface.",
)
interface_args.add_argument(
    "-p",
    "--adapter-options",
    default=[],
    action="append",
    help="Supply the protocol name and adapter options in the format "
    '"name:{opt1: val, opt2: val}". Use the -l flag to see which protocols '
    "are available for the selected device. Can be supplied multiple times for "
    "multiple protocols.",
)
device_args.add_argument(
    "-l",
    "--list-protocols",
    action="store_true",
    help="List available protocols for selected device.",
)
device_args.add_argument(
    "-L",
    "--list-adapter-options",
    action="store_true",
    help="List available configuration options and their value. Values that have not been "
    "modified in the -p argument are default values.",
)
device_args.add_argument(
    "-i",
    "--show-interface",
    action="store_true",
    help="Show command interface of device interface.",
)
device_args.add_argument(
    "-k",
    "--device-package",
    default="lewis.devices",
    help="Name of packages where devices are found.",
)
device_args.add_argument(
    "-a",
    "--add-path",
    default=None,
    help="Path where the device package exists. Is added to the path.",
)

simulation_args = parser.add_argument_group(
    "Simulation related parameters",
    "Parameters that influence the simulation itself, such as timing and speed.",
)

simulation_args.add_argument(
    "-c",
    "--cycle-delay",
    type=float,
    default=0.1,
    help="Approximate time to spend in each cycle of the simulation. "
    "0 for maximum simulation rate.",
)
simulation_args.add_argument(
    "-e",
    "--speed",
    type=float,
    default=1.0,
    help="Simulation speed. The actually elapsed time between two cycles is "
    "multiplied with this speed to determine the simulated time.",
)
simulation_args.add_argument(
    "-r",
    "--rpc-host",
    default=None,
    help="HOST:PORT format string for exposing the device and the simulation via "
    "JSON-RPC over ZMQ. Use lewis-control to access this service from the command line.",
)

other_args = parser.add_argument_group("Other arguments")

other_args.add_argument(
    "-o",
    "--output-level",
    default="info",
    choices=["none", "critical", "error", "warning", "info", "debug"],
    help="Level of detail for logging to stderr.",
)
other_args.add_argument(
    "-V",
    "--verify",
    action="store_true",
    help="Sets the output level to 'debug' and aborts before starting the device simulation. "
    "This is intended to help with diagnosing problems with devices or input arguments.",
)

version_handling = other_args.add_mutually_exclusive_group()
version_handling.add_argument(
    "-I",
    "--ignore-versions",
    action="store_true",
    help="Ignore version mismatches between device and framework. A warning will still "
    "be logged.",
)
other_args.add_argument(
    "-v", "--version", action="store_true", help="Prints the version and exits."
)
other_args.add_argument("-h", "--help", action="help", help="Shows this help message and exits.")

deprecated_args = parser.add_argument_group("Deprecated arguments")
deprecated_args.add_argument(
    "-R",
    "--relaxed-versions",
    action="store_true",
    help="Renamed to --I/--ignore-versions. Using this old option produces an error "
    "and it will be removed in a future release.",
)

__doc__ = (
    "This script is the main interaction point of the user with Lewis. The usage "
    "is as follows:\n\n.. code-block:: none\n\n{}".format(get_usage_text(parser, indent=4))
)


def parse_adapter_options(raw_adapter_options):
    if not raw_adapter_options:
        return {None: {}}

    protocols = {}

    for option_string in raw_adapter_options:
        try:
            adapter_options = yaml.safe_load(option_string)
        except yaml.YAMLError:
            raise LewisException(
                "It was not possible to parse this adapter option specification:\n"
                "    %s\n"
                "Correct formats for the -p argument are:\n"
                "    -p protocol\n"
                "    -p \"protocol: {option: 'val', option2: 34}\"\n"
                "The spaces after the colons are significant!" % option_string
            )

        if isinstance(adapter_options, str):
            protocols[adapter_options] = {}
        else:
            protocol = list(adapter_options.keys())[0]
            protocols[protocol] = adapter_options.get(protocol, {})

    return protocols


def run_simulation(argument_list=None) -> None:  # noqa: C901
    """
    This is effectively the main function of a typical simulation run. Arguments passed in are
    parsed and used to construct and run the simulation.

    This function only exits when the program has completed or is interrupted.

    :param argument_list: Argument list to pass to the argument parser declared in this module.
    """
    try:
        arguments = parser.parse_args(argument_list or sys.argv[1:])

        if arguments.version:
            print(__version__)
            return

        if arguments.relaxed_versions:
            print("Unknown option --relaxed-versions. Did you mean --ignore-versions?")
            return

        loglevel = "debug" if arguments.verify else arguments.output_level
        if loglevel != "none":
            logging.basicConfig(level=getattr(logging, loglevel.upper()), format=default_log_format)

        if arguments.add_path is not None:
            additional_path = os.path.abspath(arguments.add_path)
            logging.getLogger().debug("Extending path with: %s", additional_path)
            sys.path.append(additional_path)

        simulation_factory = SimulationFactory(arguments.device_package)

        if not arguments.device:
            devices = ["Please specify a device to simulate. The following devices are available:"]

            for dev in sorted(simulation_factory.devices):
                devices.append("    " + dev)

            print("\n".join(devices))
            return

        if arguments.list_protocols:
            print("\n".join(simulation_factory.get_protocols(arguments.device)))
            return

        protocols = (
            parse_adapter_options(arguments.adapter_options) if not arguments.no_interface else {}
        )

        simulation = simulation_factory.create(
            arguments.device, arguments.setup, protocols, arguments.rpc_host
        )

        if arguments.show_interface:
            print(simulation._adapters.documentation())
            return

        if arguments.list_adapter_options:
            configurations = simulation._adapters.configuration()

            for protocol, options in configurations.items():
                print("{}:".format(protocol))

                for opt, val in options.items():
                    print("    {} = {}".format(opt, val))

            return

        simulation.cycle_delay = arguments.cycle_delay
        simulation.speed = arguments.speed

        if not arguments.verify:
            try:
                simulation.start()
            except KeyboardInterrupt:
                print("\nInterrupt received; shutting down. Goodbye, cruel world!")
                simulation.log.critical("Simulation aborted by user interaction")
            finally:
                simulation.stop()

    except LewisException as e:
        print("\n".join(("An error occurred:", str(e))))



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest
from unittest.mock import Mock, patch

from lewis.core.statemachine import (
    State,
    StateMachine,
    StateMachineException,
    Transition,
)


class TestStateMachine(unittest.TestCase):
    def test_initial_state_is_required(self):
        with self.assertRaises(StateMachineException) as context:
            StateMachine({})

        self.assertTrue("must include 'initial'" in str(context.exception))

    def test_state_machine_starts_in_None(self):
        sm = StateMachine({"initial": "foobar"})
        self.assertIsNone(sm.state, "StateMachine should start in the 'None' non-state")

    def test_first_cycle_transitions_to_initial(self):
        sm = StateMachine({"initial": "foobar"})
        sm.process(0.1)
        self.assertEqual(
            sm.state,
            "foobar",
            "StateMachine failed to transition into " "initial state on first cycle",
        )

    def test_can_transition_with_lambda(self):
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): lambda: True}})

        self.assertIsNone(sm.state)
        sm.process(0.1)
        self.assertEqual(sm.state, "foo")
        sm.process(0.2)
        self.assertEqual(sm.state, "bar")

    def test_can_transition_with_callable(self):
        transition = Mock(return_value=True)
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): transition}})

        transition.assert_not_called()
        self.assertIsNone(sm.state)
        sm.process(0.1)
        transition.assert_not_called()
        self.assertEqual(sm.state, "foo")
        sm.process(0.2)
        self.assertTrue(transition.called)
        self.assertEqual(sm.state, "bar")

    @patch.object(Transition, "__call__", return_value=True)
    def test_can_transition_with_Transition(self, tr_call):
        transition = Transition()
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): transition}})

        tr_call.assert_not_called()
        self.assertIsNone(sm.state)
        sm.process(0.1)
        tr_call.assert_not_called()
        self.assertEqual(sm.state, "foo")
        sm.process(0.2)
        self.assertTrue(tr_call.called)
        self.assertEqual(sm.state, "bar")

    def test_Transition_receives_Context(self):
        transition = Transition()
        context = object()
        StateMachine(
            {"initial": "foo", "transitions": {("foo", "bar"): transition}},
            context=context,
        )

        self.assertEqual(transition._context, context)

    def test_can_specify_state_handlers_as_dict(self):
        on_entry = Mock()
        in_state = Mock()
        on_exit = Mock()
        sm = StateMachine(
            {
                "initial": "foo",
                "states": {
                    "foo": {
                        "on_entry": on_entry,
                        "in_state": in_state,
                        "on_exit": on_exit,
                    },
                },
                "transitions": {("foo", "bar"): lambda: True},
            }
        )

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        on_entry.assert_called_once_with(0)
        in_state.assert_called_once_with(0)

        on_entry.reset_mock()
        in_state.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        on_exit.assert_called_once_with(2.0)

        on_exit.reset_mock()

        # Third cycle only does an in_state in bar, shouldn't affect foo
        sm.process(3.0)
        on_entry.assert_not_called()
        in_state.assert_not_called()
        on_exit.assert_not_called()

    def test_can_specify_state_handlers_as_list(self):
        on_entry = Mock()
        in_state = Mock()
        on_exit = Mock()
        sm = StateMachine(
            {
                "initial": "foo",
                "states": {
                    "foo": [on_entry, in_state, on_exit],
                },
                "transitions": {("foo", "bar"): lambda: True},
            }
        )

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        on_entry.assert_called_once_with(0)
        in_state.assert_called_once_with(0)

        on_entry.reset_mock()
        in_state.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        on_exit.assert_called_once_with(2.0)

        on_exit.reset_mock()

        # Third cycle only does an in_state in bar, shouldn't affect foo
        sm.process(3.0)
        on_entry.assert_not_called()
        in_state.assert_not_called()
        on_exit.assert_not_called()

    @patch.object(State, "on_entry")
    @patch.object(State, "in_state")
    @patch.object(State, "on_exit")
    def test_can_specify_state_handlers_as_State(self, *_):
        foo = State()
        bar = State()
        sm = StateMachine(
            {
                "initial": "foo",
                "states": {
                    "foo": foo,
                    "bar": bar,
                },
                "transitions": {("foo", "bar"): lambda: True},
            }
        )

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        foo.on_entry.assert_called_once_with(0)
        foo.in_state.assert_called_once_with(0)

        foo.on_entry.reset_mock()
        foo.in_state.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        foo.on_exit.assert_called_once_with(2.0)
        bar.on_entry.assert_called_once_with(2.0)
        bar.in_state.assert_called_once_with(2.0)

        foo.on_exit.reset_mock()
        bar.on_entry.reset_mock()
        bar.in_state.reset_mock()

        # Third cycle only does an in_state in bar
        sm.process(3.0)
        bar.in_state.assert_called_once_with(3.0)
        bar.on_entry.assert_not_called()
        bar.on_exit.assert_not_called()

    def test_State_receives_Context(self):
        state = State()
        context = object()
        StateMachine({"initial": "foo", "states": {"foo": state}}, context=context)

        self.assertEqual(state._context, context)

    def test_bind_handlers_by_name_default_behaviour(self):
        target = Mock()
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): lambda: True}})
        sm.bind_handlers_by_name(target)

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        target._on_entry_foo.assert_called_once_with(0)
        target._in_state_foo.assert_called_once_with(0)

        target._on_entry_foo.reset_mock()
        target._in_state_foo.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        target._on_exit_foo.assert_called_once_with(2.0)
        target._on_entry_bar.assert_called_once_with(2.0)
        target._in_state_bar.assert_called_once_with(2.0)

        target._on_exit_foo.reset_mock()
        target._on_entry_bar.reset_mock()
        target._in_state_bar.reset_mock()

        # Third cycle only does an in_state in bar
        sm.process(3.0)
        target._in_state_bar.assert_called_once_with(3.0)
        target._on_entry_bar.assert_not_called()
        target._on_exit_bar.assert_not_called()

    def test_bind_handlers_by_name_custom_prefix(self):
        target = Mock()
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): lambda: True}})
        sm.bind_handlers_by_name(
            target,
            prefix={
                "on_entry": "enter_",
                "in_state": "do_",
                "on_exit": "exit_",
            },
        )

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        target.enter_foo.assert_called_once_with(0)
        target.do_foo.assert_called_once_with(0)

        target.enter_foo.reset_mock()
        target.do_foo.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        target.exit_foo.assert_called_once_with(2.0)
        target.enter_bar.assert_called_once_with(2.0)
        target.do_bar.assert_called_once_with(2.0)

        target.exit_foo.reset_mock()
        target.enter_bar.reset_mock()
        target.do_bar.reset_mock()

        # Third cycle only does an in_state in bar
        sm.process(3.0)
        target.do_bar.assert_called_once_with(3.0)
        target.enter_bar.assert_not_called()
        target.exit_bar.assert_not_called()

    def test_bind_handlers_by_name_override(self):
        # first target will accept any event call
        first = Mock()

        # second target will 'override' the events for bar
        second = Mock(spec=["_on_entry_bar", "_in_state_bar", "_on_exit_bar"])

        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): lambda: True}})
        sm.bind_handlers_by_name(first)
        sm.bind_handlers_by_name(second, override=True)

        # First cycle enters and executes initial state, but forces delta T to zero
        sm.process(1.0)
        first._on_entry_foo.assert_called_once_with(0)
        first._in_state_foo.assert_called_once_with(0)

        first._on_entry_foo.reset_mock()
        first._in_state_foo.reset_mock()

        # Second cycle transitions due to lambda: True above
        sm.process(2.0)
        first._on_exit_foo.assert_called_once_with(2.0)
        second._on_entry_bar.assert_called_once_with(2.0)
        second._in_state_bar.assert_called_once_with(2.0)

        first._on_exit_foo.reset_mock()
        second._on_entry_bar.reset_mock()
        second._in_state_bar.reset_mock()

        # Third cycle only does an in_state in bar
        sm.process(3.0)
        second._in_state_bar.assert_called_once_with(3.0)
        second._on_entry_bar.assert_not_called()
        second._on_exit_bar.assert_not_called()

        # 'bar' events should not have been called on 'first'
        first._on_entry_bar.assert_not_called()
        first._in_state_bar.assert_not_called()
        first._on_exit_bar.assert_not_called()

    def test_reset(self):
        sm = StateMachine({"initial": "foo", "transitions": {("foo", "bar"): lambda: True}})

        self.assertIsNone(sm.state)
        sm.reset()
        self.assertIsNone(sm.state)

        sm.process(0.1)
        self.assertEqual(sm.state, "foo")
        sm.reset()
        self.assertIsNone(sm.state)

        sm.process(0.2)
        sm.process(0.3)
        self.assertEqual(sm.state, "bar")
        sm.reset()
        self.assertIsNone(sm.state)

    def test_can(self):
        sm = StateMachine(
            {
                "initial": "init",
                "transitions": {
                    ("init", "foo"): lambda: True,
                    ("foo", "bar"): lambda: True,
                    ("bar", "foo"): lambda: True,
                    ("bar", "init"): lambda: True,
                    ("bar", "bar"): lambda: True,
                },
            }
        )

        self.assertEqual(sm.state, None)
        self.assertIs(sm.can("init"), True)
        self.assertIs(sm.can("foo"), False)
        self.assertIs(sm.can("bar"), False)
        self.assertIs(sm.can(None), False)

        sm.process()
        self.assertEqual(sm.state, "init")
        self.assertIs(sm.can("foo"), True)
        self.assertIs(sm.can("bar"), False)
        self.assertIs(sm.can("init"), False)

        sm.process()
        self.assertEqual(sm.state, "foo")
        self.assertIs(sm.can("bar"), True)
        self.assertIs(sm.can("init"), False)
        self.assertIs(sm.can("foo"), False)

        sm.process()
        self.assertEqual(sm.state, "bar")
        self.assertIs(sm.can("foo"), True)
        self.assertIs(sm.can("init"), True)
        self.assertIs(sm.can("bar"), True)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import inspect
from datetime import datetime
from functools import wraps

from lewis.core.adapters import Adapter
from lewis.core.devices import InterfaceBase
from lewis.core.exceptions import (
    AccessViolationException,
    LewisException,
    LimitViolationException,
)
from lewis.core.logging import has_log
from lewis.core.utils import FromOptionalDependency, format_doc_text, seconds_since

# pcaspy might not be available. To make EPICS-based adapters show up
# in the listed adapters anyway dummy types are created in this case
# and the failure is postponed to runtime, where a more appropriate
# LewisException can be raised.
missing_pcaspy_exception = LewisException(
    "In order to use EPICS-interfaces, pcaspy must be installed:\n"
    "\tpip install pcaspy\n"
    "A fully working installation of EPICS-base is required for this package. "
    "Please refer to the documentation for advice."
)

Driver, SimpleServer = FromOptionalDependency("pcaspy", missing_pcaspy_exception).do_import(
    "Driver", "SimpleServer"
)

pcaspy_manager = FromOptionalDependency("pcaspy.driver", missing_pcaspy_exception).do_import(
    "manager"
)


class BoundPV:
    """
    Class to represent PVs that are bound to an adapter

    This class is very similar to :class:`~lewis.adapters.stream.Func`, in that
    it is the result of a binding operation between a user-specified :class:`PV`-object
    and a Device and/or Adapter object. Also, it should rarely be used directly. objects
    are generated automatically by :class:`EpicsAdapter`.

    The binding happens by supplying a ``target``-object which has an attribute or a property
    named according to the property-name stored in the PV-object, and a ``meta_target``-object
    which has an attribute named according to the meta_data_property in PV.

    The properties ``read_only``, ``config``,  and ``poll_interval`` simply forward the
    data of PV, while ``doc`` uses the target object to potentially obtain the property's
    docstring.

    To get and set the value of the property on the target, the ``value``-property of
    this class can be used, to get the meta data dict, there's a ``meta``-property.

    :param pv: PV object to bind to target and meta_target.
    :param target: Object that has an attribute named pv.property.
    :param meta_target: Object that has an attribute named pv.meta_data_property.
    """

    def __init__(self, pv, target, meta_target=None) -> None:
        self._meta_target = meta_target
        self._target = target
        self._pv = pv

    @property
    def value(self):
        """Value of the bound property on the target."""
        return getattr(self._target, self._pv.property)

    @value.setter
    def value(self, new_value) -> None:
        if self.read_only:
            raise AccessViolationException(
                "The property {} is read only.".format(self._pv.property)
            )

        setattr(self._target, self._pv.property, new_value)

    @property
    def meta(self):
        """Value of the bound meta-property on the target."""
        if not self._pv.meta_data_property or not self._meta_target:
            return {}

        return getattr(self._meta_target, self._pv.meta_data_property)

    @property
    def read_only(self):
        """True if the PV is read-only."""
        return self._pv.read_only

    @property
    def config(self):
        """Config dict passed on to pcaspy-machinery."""
        return self._pv.config

    @property
    def poll_interval(self):
        """Interval at which to update PV in pcaspy."""
        return self._pv.poll_interval

    @property
    def doc(self):
        """Docstring of property on target or override specified on PV-object."""
        return (
            self._pv.doc
            or inspect.getdoc(getattr(type(self._target), self._pv.property, None))
            or ""
        )


class PV:
    """
    The PV-class is used to declare the EPICS-interface exposed by a sub-class of
    EpicsAdapter. The ``target_property`` argument specifies which property of the adapter
    the PV maps to. To make development easier it can also be a part of the exposed
    device. If the property exists on both the Adapter-subclass and the device, the former
    has precedence. This is useful for overriding behavior for protocol specific "quirks".

    If the PV should be read only, this needs to be specified via
    the corresponding parameter. The information about the poll interval is used
    py EpicsAdapter to update the PV in regular intervals. All other named arguments
    are forwarded to the pcaspy server's `pvdb`, so it's possible to pass on
    limits, types, enum-values and so on.

    In case those arguments change at runtime, it's possible to provide ``meta_data_property``,
    which should contain the name of a property that returns a dict containing these values.
    For example if limits change:

    .. sourcecode:: Python

        class Interface(EpicsInterface):
            pvs = {"example": PV("example", meta_data_property="example_meta")}

            @property
            def example_meta(self):
                return {
                    "lolim": self.device._example_low_limit,
                    "hilim": self.device._example_high_limit,
                }

    The PV infos are then updated together with the value, determined by ``poll_interval``.

    In cases where the device is accessed via properties alone, this class provides the possibility
    to expose methods as PVs. A common use case would be to model a getter:

    .. sourcecode:: Python

        class SomeDevice(Device):
            def get_example(self):
                return 42


        class Interface(EpicsInterface):
            pvs = {"example": PV("get_example")}

    It is also possible to model a getter/setter pair, in this case a tuple has to be provided:

    .. sourcecode:: Python

        class SomeDevice(Device):
            _ex = 40

            def get_example(self):
                return self._ex + 2

            def set_example(self, new_example):
                self._ex = new_example - 2


        class Interface(EpicsInterface):
            pvs = {"example": PV(("get_example", "set_example"))}

    Any of the two members in the tuple can be substituted with ``None`` in case it does not apply.
    Besides method names it is also allowed to provide callables. Valid callables are for example
    bound methods and free functions, but also lambda expressions and partials.

    There are however restrictions for the supplied functions (be it as method names or directly
    as callables) with respect to their signature. Getter functions must be callable without any
    arguments, setter functions must be callable with exactly one argument. The ``self`` of
    methods does not count towards this.


    :param target_property: Property or method name, getter function, tuple of getter/setter.
    :param poll_interval: Update interval of the PV.
    :param read_only: Should be True if the PV is read only. If not specified, the PV is
                      read_only if only a getter is supplied.
    :param meta_data_property: Property or method name, getter function, tuple of getter/setter.
    :param doc: Description of the PV. If not supplied, docstring of mapped property is used.
    :param kwargs: Arguments forwarded into pcaspy pvdb-dict.
    """

    def __init__(
        self,
        target_property,
        poll_interval=1.0,
        read_only=False,
        meta_data_property=None,
        doc=None,
        **kwargs,
    ) -> None:
        self.property = "value"
        self.read_only = read_only
        self.poll_interval = poll_interval
        self.meta_data_property = "meta"
        self.doc = doc
        self.config = kwargs

        value = self._get_specification(target_property)
        meta = self._get_specification(meta_data_property)

        self._specifications = {"value": value, "meta": meta}

    def bind(self, *targets):
        """
        Tries to bind the PV to one of the supplied targets. Targets are inspected according to
        the order in which they are supplied.

        :param targets: Objects to inspect from.
        :return: BoundPV instance with the PV bound to the target property.
        """
        self.property = "value"
        self.meta_data_property = "meta"

        return BoundPV(
            self,
            self._get_target(self.property, *targets),
            self._get_target(self.meta_data_property, *targets),
        )

    def _get_specification(self, spec):
        """
        Helper method to create a homogeneous representation of a specified getter or
        getter/setter pair.

        :param spec: Function specification 'getter', (getter,) or (getter,setter)
        :return:  Harmonized getter/setter specification, (getter, setter)
        """
        if spec is None or callable(spec) or isinstance(spec, str):
            spec = (spec,)
        if len(spec) == 1:
            spec = (spec[0], None)
        return spec

    def _get_target(self, prop, *targets):
        """
        The actual target methods are retrieved (possibly from the list of targets) and a
        wrapper-property is installed on a throwaway type that is specifically created for
        the purpose of holding this property if necessary. In that case, an instance of this type
        (with the wrapper-property forwarding calls to the correct target) is returned so
        that :class:`BoundPV` can do the right thing.

        .. seealso:: :meth:`_create_getter`, :meth:`_create_setter`

        :param prop: Property, is either 'value' or 'meta'.
        :param targets: List of targets with decreasing priority for finding the wrapped method.
        :return: Target object to be used by :class:`BoundPV`.
        """

        if prop is None:
            return None

        raw_getter, raw_setter = self._specifications.get(prop, (None, None))

        target = None

        if isinstance(raw_getter, str):
            target = next(
                (
                    obj
                    for obj in targets
                    if isinstance(getattr(type(obj), raw_getter, None), property)
                    or not callable(getattr(obj, raw_getter, lambda: True))
                ),
                None,
            )

        if target is not None:
            # If the property is an actual property and has no setter, read_only can be
            # set to True at this point automatically.
            target_prop = getattr(type(target), raw_getter, None)

            if prop == "value" and isinstance(target_prop, property) and target_prop.fset is None:
                self.read_only = True

            # Now the target does not need to be constructed, property or meta_data_property
            # needs to change.
            setattr(
                self,
                "property" if prop == "value" else "meta_data_property",
                raw_getter,
            )
            return target

        getter = self._create_getter(raw_getter, *targets)
        setter = self._create_setter(raw_setter, *targets)

        if getter is None and setter is None:
            return None

        if prop == "value" and setter is None:
            self.read_only = True

        return type(prop, (object,), {prop: property(getter, setter)})()

    def _create_getter(self, func, *targets):
        """
        Returns a function wrapping the supplied function. The returned wrapper can be used as the
        getter in a property definition. Raises a RuntimeError if the signature of the supplied
        function is not compatible with the getter-concept (no arguments except self).

        :param func: Callable or name of method on one object in targets.
        :param targets: List of targets with decreasing priority for finding func.
        :return: Getter function for constructing a wrapper-property.
        """
        if not func:
            return None

        final_callable = self._get_callable(func, *targets)

        if not self._function_has_n_args(final_callable, 0):
            raise RuntimeError(
                "The function '{}' does not look like a getter function. A valid getter "
                "function has no arguments that do not have a default. The self-argument of "
                "methods does not count towards that number.".format(final_callable.__name__)
            )

        @wraps(final_callable)
        def getter(obj):
            return final_callable()

        return getter

    def _create_setter(self, func, *targets):
        """
        Returns a function wrapping the supplied function. The returned wrapper can be used as the
        setter in a property definition. Raises a RuntimeError if the signature of the supplied
        function is not compatible with the setter-concept (exactly one argument except self).

        :param func: Callable or name of method on one object in targets.
        :param targets: List of targets with decreasing priority for finding func.
        :return: Setter function for constructing a wrapper-property or ``None``.
        """
        if not func:
            return None

        func = self._get_callable(func, *targets)

        if not self._function_has_n_args(func, 1):
            raise RuntimeError(
                "The function '{}' does not look like a setter function. A valid setter "
                "function has exactly one argument without a default. The self-argument of "
                "methods does not count towards that number.".format(func.__name__)
            )

        def setter(obj, value) -> None:
            func(value)

        return setter

    def _get_callable(self, func, *targets):
        """
        If func is already a callable, it is returned directly. If it's a string, it is assumed
        to be a method on one of the objects supplied in targets and that is returned. If no
        method with the specified name is found, an AttributeError is raised.

        :param func: Callable or name of method on one object in targets.
        :param targets: List of targets with decreasing priority for finding func.
        :return: Callable.
        """
        if not callable(func):
            func_name = func
            func = next((getattr(obj, func, None) for obj in targets if func in dir(obj)), None)

            if not func:
                raise AttributeError(
                    "No method with the name '{}' could be found on any of the target objects "
                    "(device, interface). Please check the spelling.".format(func_name)
                )

        return func

    def _function_has_n_args(self, func, n):
        """
        Returns true if func has n arguments. Arguments with default and self for
        methods are not considered.
        """
        if inspect.ismethod(func):
            n += 1

        argspec = inspect.getargspec(func)
        defaults = argspec.defaults or ()

        return len(argspec.args) - len(defaults) == n


@has_log
class PropertyExposingDriver(Driver):
    def __init__(self, interface, device_lock) -> None:
        super(PropertyExposingDriver, self).__init__()

        self._interface = interface
        self._device_lock = device_lock
        self._set_logging_context(interface)

        self._timers = {k: 0.0 for k in self._interface.bound_pvs.keys()}
        self._last_update_call = None

    def write(self, pv, value) -> bool:
        self.log.debug("PV put request: %s=%s", pv, value)

        pv_object = self._interface.bound_pvs.get(pv)

        if not pv_object:
            return False

        try:
            with self._device_lock:
                pv_object.value = value
                self.setParam(pv, pv_object.value)

                return True
        except LimitViolationException as e:
            self.log.warning(
                "Rejected writing value %s to PV %s due to limit " "violation. %s",
                value,
                pv,
                e,
            )
        except AccessViolationException:
            self.log.warning(
                "Rejected writing value %s to PV %s due to access " "violation, PV is read-only.",
                value,
                pv,
            )

        return False

    def _get_param_info(self, pv, meta_keys):
        """
        Get PV info fields from pcaspy's "manager" object. This function returns a dictionary
        with info/value pairs, where each entry of meta_keys results in a dictionary entry if
        pcaspy's PVInfo-object has such an attribute. Attributes that do not exist are ignored.
        Valid attributes are the same as specified in the ``pvdb``-argument that

        :param pv: PV base name
        :param meta_keys: List of keys for what information to obtain
        :return:
        """
        # TODO: Submit upstream patch to make this method available in base class
        pv = pcaspy_manager.pvs[self.port][pv]

        info_dict = {}
        for key in meta_keys:
            if hasattr(pv.info, key):
                info_dict[key] = getattr(pv.info, key)

        return info_dict

    def process_pv_updates(self, force=False) -> None:
        """
        Update PV values that have changed for PVs that are due to update according to their
        respective poll interval timers.

        :param force: If True, will force updates to all PVs regardless of timers.
        """
        dt = seconds_since(self._last_update_call or datetime.now())

        # Cache details of PVs that need to update
        value_updates = []
        meta_updates = []

        with self._device_lock:
            for pv, pv_object in self._interface.bound_pvs.items():
                self._timers[pv] = self._timers.get(pv, 0.0) + dt
                if self._timers[pv] >= pv_object.poll_interval or force:
                    try:
                        if self.getParam(pv) != pv_object.value or force:
                            value_updates.append((pv, pv_object.value))

                        pv_meta = pv_object.meta
                        if self._get_param_info(pv, pv_meta.keys()) != pv_meta or force:
                            meta_updates.append((pv, pv_meta))

                    except (AttributeError, TypeError):
                        self.log.exception("An error occurred while updating PV %s.", pv)
                    finally:
                        self._timers[pv] = 0.0

        self._process_value_updates(value_updates)
        self._process_meta_updates(meta_updates)

        self._last_update_call = datetime.now()

    def _process_value_updates(self, updates) -> None:
        if updates:
            update_log = []
            for pv, value in updates:
                self.setParam(pv, value)
                update_log.append("{}={}".format(pv, value))

            self.log.info("Processed PV updates: %s", ", ".join(update_log))

            # Calling this manually is only required for values, not for meta
            self.updatePVs()

    def _process_meta_updates(self, updates) -> None:
        if updates:
            update_log = []
            for pv, info in updates:
                self.setParamInfo(pv, info)
                update_log.append("{}={}".format(pv, info))

            self.log.info("Processed PV-info updates: %s", ", ".join(update_log))


class EpicsAdapter(Adapter):
    """
    This adapter provides ChannelAccess server functionality through the pcaspy module.

    It's possible to configure the prefix for the PVs provided by this adapter. The
    corresponding key in the ``options`` dictionary is called ``prefix``:

    .. sourcecode:: Python

        options = {"prefix": "PVPREFIX:"}

    :param options: Dictionary with options.
    """

    default_options = {"prefix": ""}

    def __init__(self, options=None) -> None:
        super(EpicsAdapter, self).__init__(options)

        self._server = None
        self._driver = None

    @property
    def documentation(self):
        pvs = []

        for name, pv in self.interface.bound_pvs.items():
            complete_name = self._options.prefix + name

            data_type = pv.config.get("type", "float")
            read_only_tag = ", read only" if pv.read_only else ""

            pvs.append(
                "{} ({}{}):\n{}".format(
                    complete_name, data_type, read_only_tag, format_doc_text(pv.doc)
                )
            )

        return "\n\n".join([inspect.getdoc(self.interface) or "", "PVs\n==="] + pvs)

    def start_server(self) -> None:
        """
        Creates a pcaspy-server.

        .. note::

            The server does not process requests unless :meth:`handle` is called regularly.
        """
        if self._server is None:
            self._server = SimpleServer()
            self._server.createPV(
                prefix=self._options.prefix,
                pvdb={k: v.config for k, v in self.interface.bound_pvs.items()},
            )
            self._driver = PropertyExposingDriver(
                interface=self.interface, device_lock=self.device_lock
            )
            self._driver.process_pv_updates(force=True)

            self.log.info(
                "Started serving PVs: %s",
                ", ".join((self._options.prefix + pv for pv in self.interface.bound_pvs.keys())),
            )

    def stop_server(self) -> None:
        self._driver = None
        self._server = None

    @property
    def is_running(self):
        return self._server is not None

    def handle(self, cycle_delay=0.1) -> None:
        """
        Call this method to spend about ``cycle_delay`` seconds processing
        requests in the pcaspy server. Under load, for example when running ``caget`` at a
        high frequency, the actual time spent in the method may be much shorter. This effect
        is not corrected for.

        :param cycle_delay: Approximate time to be spent processing requests in pcaspy server.
        """
        if self._server is not None:
            self._server.process(cycle_delay)
            self._driver.process_pv_updates()


class EpicsInterface(InterfaceBase):
    """
    Inheriting from this class provides an EPICS-interface to a device for use with
    :class:`EpicsAdapter`. In the simplest case all that is required is to inherit
    from this class and override the ``pvs``-member. It should be a dictionary
    that contains PV-names (without prefix) as keys and instances of PV as
    values. The prefix is handled by ``EpicsAdapter``.

    For a simple device with two properties, speed and position, the first of which
    should be read-only, it's enough to define the following:

    .. sourcecode:: Python

        class SimpleDeviceEpicsInterface(EpicsInterface):
            pvs = {"VELO": PV("speed", read_only=True), "POS": PV("position", lolo=0, hihi=100)}

    For more complex behavior, the interface could contain properties that do not
    exist in the device itself. If the device should also have a PV called STOP
    that "stops the device", the interface could look like this:

    .. sourcecode:: Python

        class SimpleDeviceEpicsInterface(EpicsInterface):
            pvs = {
                "VELO": PV("speed", read_only=True),
                "POS": PV("position", lolo=0, hihi=100),
                "STOP": PV("stop", type="int"),
            }

            @property
            def stop(self):
                return 0

            @stop.setter
            def stop(self, value):
                if value == 1:
                    self.device.halt()

    Even though the device does *not* have a property called ``stop`` (but a method called
    ``halt``), issuing the command

    ::

        $ caput STOP 1

    will achieve the desired behavior, because ``EpicsInterface`` merges the properties
    of the device into ``SimpleDeviceEpicsInterface`` itself, so that it is does not
    matter whether the specified property in PV exists in the device or the adapter.

    The intention of this design is to keep device classes small and free of
    protocol specific stuff, such as in the case above where stopping a device
    via EPICS might involve writing a value to a PV, whereas other protocols may
    offer an RPC-way of achieving the same thing.
    """

    protocol = "epics"
    pvs = None

    def __init__(self) -> None:
        super(EpicsInterface, self).__init__()
        self.bound_pvs = None

    @property
    def adapter(self):
        return EpicsAdapter

    def _bind_device(self) -> None:
        """
        This method transforms the ``self.pvs`` dict of :class:`PV` objects ``self.bound_pvs``,
        a dict of :class:`BoundPV` objects, the keys are always the PV-names that are exposed
        via ChannelAccess.

        In the transformation process, the method tries to find whether the attribute specified by
        PV's ``property`` (and ``meta_data_property``) is part of the internally stored device
        or the interface and constructs a BoundPV, which acts as a forwarder to the appropriate
        objects.
        """
        self.bound_pvs = {}

        for pv_name, pv in self.pvs.items():
            try:
                self.bound_pvs[pv_name] = pv.bind(self, self.device)
            except (AttributeError, RuntimeError) as e:
                self.log.debug(
                    "An exception was caught during the binding step of PV '%s'.",
                    pv_name,
                    exc_info=e,
                )
                raise LewisException(
                    "The binding step for PV '{}' failed, please check the interface-"
                    "definition or contact the device author. More information is "
                    "available with debug-level logging (-o debug).".format(pv_name)
                )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import importlib
import unittest
from datetime import datetime
from unittest.mock import patch

from lewis.core.exceptions import LewisException, LimitViolationException
from lewis.core.utils import (
    FromOptionalDependency,
    check_limits,
    dict_strict_update,
    extract_module_name,
    format_doc_text,
    get_members,
    get_submodules,
    seconds_since,
)

from .utils import TestWithPackageStructure, assertRaisesNothing


class TestDictStrictUpdate(unittest.TestCase):
    def test_update_all(self):
        base_dict = {1: 2, 2: 45, 4: 4}
        update_dict = {1: 3, 2: 43, 4: 3}

        dict_strict_update(base_dict, update_dict)

        self.assertEqual(base_dict, update_dict)

    def test_update_subset(self):
        base_dict = {1: 2, 2: 45, 4: 4}
        update_dict = {1: 3, 2: 43}

        dict_strict_update(base_dict, update_dict)

        self.assertEqual(base_dict, {1: 3, 2: 43, 4: 4})

    def test_update_superset_exception(self):
        base_dict = {1: 2, 2: 45, 4: 4}
        update_dict = {51: 3, 2: 43}

        self.assertRaises(RuntimeError, dict_strict_update, base_dict, update_dict)


class TestExtractModuleName(TestWithPackageStructure):
    def test_directory_basename_is_returned(self):
        self.assertEqual(extract_module_name(self._dirs["valid"]), "some_dir")

    def test_directory_invalid_name(self):
        self.assertEqual(extract_module_name(self._dirs["invalid_underscore"]), None)
        self.assertEqual(extract_module_name(self._dirs["invalid_dot"]), None)

    def test_file_invalid_name(self):
        self.assertEqual(extract_module_name(self._files["invalid_name"]), None)

    def test_file_invalid_extension(self):
        self.assertEqual(extract_module_name(self._files["invalid_ext"]), None)

    def test_file_basename_without_extension(self):
        self.assertEqual(extract_module_name(self._files["valid"]), "some_file")


class TestGetSubmodules(TestWithPackageStructure):
    def test_non_module_raises_runtimeerror(self):
        self.assertRaises(RuntimeError, get_submodules, self._tmp_package_name)

    def test_correct_modules_are_returned(self):
        submodules = get_submodules(importlib.import_module(self._tmp_package_name))

        self.assertEqual(sorted(submodules.keys()), sorted(self._expected_modules))


class TestGetMembers(unittest.TestCase):
    def test_returns_all_members_if_predicate_is_missing(self):
        class Foo:
            bar = 3.0
            baz = "test"

        members = get_members(Foo())

        self.assertEqual(len(members), 2)
        self.assertIn("bar", members)
        self.assertIn("baz", members)

    def test_predicate(self):
        class Foo:
            bar = 3.0
            baz = "test"

        members = get_members(Foo(), lambda x: isinstance(x, str))

        self.assertEqual(len(members), 1)
        self.assertIn("baz", members)


class TestSecondsSince(unittest.TestCase):
    @patch("lewis.core.utils.datetime")
    def test_seconds_since_past(self, datetime_mock):
        datetime_mock.now.return_value = datetime(2016, 9, 1, 2, 0)

        self.assertEqual(seconds_since(datetime(2016, 9, 1, 1, 0)), 3600.0)

    @patch("lewis.core.utils.datetime")
    def test_seconds_since_future(self, datetime_mock):
        datetime_mock.now.return_value = datetime(2016, 9, 1, 2, 0)

        self.assertEqual(seconds_since(datetime(2016, 9, 1, 3, 0)), -3600.0)

    @patch("lewis.core.utils.datetime")
    def test_seconds_since_none(self, datetime_mock):
        datetime_mock.now.return_value = datetime(2016, 9, 1, 2, 0)

        self.assertRaises(TypeError, seconds_since, None)


class TestFromOptionalDependency(unittest.TestCase):
    def test_existing_module_works(self):
        a = FromOptionalDependency("time").do_import("sleep")

        from time import sleep as b

        self.assertEqual(a, b)

    def test_non_existing_members_in_module_dont_work(self):
        self.assertRaises(
            AttributeError, FromOptionalDependency("time").do_import, "sleep", "bleep"
        )

    def test_non_existing_module_works(self):
        a, b = FromOptionalDependency("invalid_module").do_import("A", "B")

        self.assertEqual(a.__name__, "A")
        self.assertEqual(b.__name__, "B")

        self.assertRaises(LewisException, a, "argument_one")
        self.assertRaises(LewisException, b, "argument_one", "argument_two")

    def test_string_exception_is_raised(self):
        a = FromOptionalDependency("invalid_module", "test").do_import("A")

        self.assertRaises(LewisException, a)

    def test_custom_exception_is_raised(self):
        a = FromOptionalDependency("invalid_module", ValueError("test")).do_import("A")

        self.assertRaises(ValueError, a)

    def test_exception_does_not_accept_arbitrary_type(self):
        self.assertRaises(RuntimeError, FromOptionalDependency, "invalid_module", 6.0)


class TestFormatDocText(unittest.TestCase):
    def test_lines_are_preserved_and_indented(self):
        text = "This is\na test\nwith multiple lines."
        expected = "    This is\n    a test\n    with multiple lines."
        self.assertEqual(format_doc_text(text), expected)

    def test_indented_lines_are_cleaned_up(self):
        text = "  This is\n  a test\n  with multiple lines."
        expected = "    This is\n    a test\n    with multiple lines."
        self.assertEqual(format_doc_text(text), expected)

    def test_long_lines_are_broken(self):
        text = " ".join(["ab"] * 44)
        expected = "    " + " ".join(["ab"] * 32) + "\n" + "    " + " ".join(["ab"] * 12)

        self.assertEqual(format_doc_text(text), expected)

    def test_no_lines_above_99(self):
        text = " ".join(["abc"] * 143)
        converted = format_doc_text(text).split("\n")

        self.assertTrue(all([len(line) <= 99 for line in converted]))


class TestCheckLimits(unittest.TestCase):
    def test_static_limits(self):
        class Foo:
            bar = 0

            @check_limits(0, 15)
            def set_bar(self, new_bar):
                self.bar = new_bar

        f = Foo()

        assertRaisesNothing(self, f.set_bar, 0)
        assertRaisesNothing(self, f.set_bar, 15)
        assertRaisesNothing(self, f.set_bar, 7)

        self.assertRaises(LimitViolationException, f.set_bar, -3)
        self.assertRaises(LimitViolationException, f.set_bar, 16)

    def test_upper_lower_only(self):
        class Foo:
            bar = 0
            baz = 1

            @check_limits(upper=15)
            def set_bar(self, new_bar):
                self.bar = new_bar

            @check_limits(lower=0)
            def set_baz(self, new_baz):
                self.baz = new_baz

        f = Foo()

        assertRaisesNothing(self, f.set_bar, 0)
        assertRaisesNothing(self, f.set_bar, 15)
        assertRaisesNothing(self, f.set_bar, -5)
        self.assertRaises(LimitViolationException, f.set_bar, 16)

        assertRaisesNothing(self, f.set_baz, 0)
        assertRaisesNothing(self, f.set_baz, 15)
        assertRaisesNothing(self, f.set_baz, 16)
        self.assertRaises(LimitViolationException, f.set_baz, -5)

    def test_property_limits(self):
        class Foo:
            bar = 0
            bar_min = 0
            bar_max = 15

            @check_limits("bar_min", "bar_max")
            def set_bar(self, new_bar):
                self.bar = new_bar

        f = Foo()

        assertRaisesNothing(self, f.set_bar, 0)
        assertRaisesNothing(self, f.set_bar, 15)

        self.assertRaises(LimitViolationException, f.set_bar, -3)
        self.assertRaises(LimitViolationException, f.set_bar, 16)

        f.bar_min = -3
        f.bar_max = 16

        assertRaisesNothing(self, f.set_bar, -3)
        assertRaisesNothing(self, f.set_bar, 16)

        f.bar_min = None
        f.bar_max = None

        assertRaisesNothing(self, f.set_bar, 123232224)
        assertRaisesNothing(self, f.set_bar, -352622234)

    def test_silent_mode(self):
        class Foo:
            bar = 0

            @check_limits(0, 15, silent=True)
            def set_bar(self, new_bar):
                self.bar = new_bar

        f = Foo()

        assertRaisesNothing(self, f.set_bar, 0)
        assertRaisesNothing(self, f.set_bar, 15)

        assertRaisesNothing(self, f.set_bar, -3)
        assertRaisesNothing(self, f.set_bar, 16)

        # Updates must have been ignored.
        self.assertEqual(f.bar, 15)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.stream import Cmd, StreamInterface, Var
from lewis.devices import Device


class VerySimpleDevice(Device):
    param = 10


class VerySimpleInterface(StreamInterface):
    """
    A very simple device with TCP-stream interface

    The device has only one parameter, which can be set to an arbitrary
    value. The interface consists of five commands which can be invoked via telnet.
    To connect:

        $ telnet host port

    After that, typing either of the commands and pressing enter sends them to the server.

    The commands are:

     - ``V``: Returns the parameter as part of a verbose message.
     - ``V=something``: Sets the parameter to ``something``.
     - ``P``: Returns the device parameter unmodified.
     - ``P=something``: Exactly the same as ``V=something``.
     - ``R`` or ``r``: Returns the number 4.

    """

    commands = {
        Cmd("get_param", pattern="^V$", return_mapping="The value is {}".format),
        Cmd("set_param", pattern="^V=(.+)$", argument_mappings=(int,)),
        Var(
            "param",
            read_pattern="^P$",
            write_pattern="^P=(.+)$",
            doc="The only parameter.",
        ),
        Cmd(lambda: 4, pattern="^R$(?i)", doc='"Random" number (4).'),
    }

    in_terminator = "\r\n"
    out_terminator = "\r\n"

    def get_param(self):
        """Returns the device parameter."""
        return self.device.param

    def set_param(self, new_param) -> None:
        """Set the device parameter, does not return anything."""
        self.device.param = new_param

    def handle_error(self, request, error):
        return "An error occurred: " + repr(error)



import inspect
import unittest
from unittest.mock import MagicMock, Mock

from lewis.core.adapters import Adapter, AdapterCollection, NoLock
from lewis.core.exceptions import LewisException

from .utils import assertRaisesNothing


class DummyAdapter(Adapter):
    """
    A dummy adapter for tests.
    """

    default_options = {
        "foo": True,
        "bar": False,
    }

    def __init__(self, protocol, running=False, options=None):
        super(DummyAdapter, self).__init__(options)
        self._protocol = protocol
        self._running = running

    @property
    def protocol(self):
        return self._protocol

    def start_server(self):
        self._running = True

    def stop_server(self):
        self._running = False

    @property
    def is_running(self):
        return self._running


class TestNoLock(unittest.TestCase):
    def test_raises_when_used(self):
        def failing_function():
            with NoLock():
                pass

        self.assertRaises(RuntimeError, failing_function)


class TestAdapter(unittest.TestCase):
    def test_documentation(self):
        adapter = DummyAdapter("foo")

        self.assertEqual(inspect.cleandoc(adapter.__doc__), adapter.documentation)

    def test_not_implemented_errors(self):
        adapter = Adapter()

        self.assertRaises(NotImplementedError, adapter.start_server)
        self.assertRaises(NotImplementedError, adapter.stop_server)
        self.assertRaises(NotImplementedError, getattr, adapter, "is_running")
        assertRaisesNothing(self, adapter.handle, 0)

    def test_interface_property(self):
        adapter = Adapter()
        mock_interface = Mock()

        # Make sure that the default implementation works (for adapters that do
        # not have binding behavior).
        adapter.interface = mock_interface
        self.assertEqual(adapter.interface, mock_interface)

    def test_protocol_is_forwarded_from_interface(self):
        adapter = Adapter()

        adapter.interface = None
        self.assertEqual(adapter.protocol, None)

        mock_interface = MagicMock()
        mock_interface.protocol = "foo"

        adapter.interface = mock_interface

        self.assertEqual(adapter.protocol, "foo")

    def test_options(self):
        assertRaisesNothing(self, DummyAdapter, "protocol", options={"bar": 2, "foo": 3})
        self.assertRaises(LewisException, DummyAdapter, "protocol", options={"invalid": False})


class TestAdapterCollection(unittest.TestCase):
    def test_add_adapter(self):
        collection = AdapterCollection()
        self.assertEqual(len(collection.protocols), 0)

        assertRaisesNothing(self, collection.add_adapter, DummyAdapter("foo"))

        self.assertEqual(len(collection.protocols), 1)
        self.assertSetEqual(set(collection.protocols), {"foo"})

        assertRaisesNothing(self, collection.add_adapter, DummyAdapter("bar"))

        self.assertEqual(len(collection.protocols), 2)
        self.assertSetEqual(set(collection.protocols), {"foo", "bar"})

        self.assertRaises(RuntimeError, collection.add_adapter, DummyAdapter("bar"))

    def test_remove_adapter(self):
        collection = AdapterCollection(DummyAdapter("foo"))

        self.assertSetEqual(set(collection.protocols), {"foo"})
        self.assertRaises(RuntimeError, collection.remove_adapter, "bar")

        assertRaisesNothing(self, collection.remove_adapter, "foo")

        self.assertEqual(len(collection.protocols), 0)

    def test_connect_disconnect_connected(self):
        collection = AdapterCollection(
            DummyAdapter("foo", running=False), DummyAdapter("bar", running=False)
        )

        # no arguments connects everything
        collection.connect()

        self.assertDictEqual(collection.is_connected(), {"bar": True, "foo": True})
        self.assertTrue(collection.is_connected("bar"))
        self.assertTrue(collection.is_connected("foo"))

        collection.disconnect()

        self.assertDictEqual(collection.is_connected(), {"bar": False, "foo": False})
        self.assertFalse(collection.is_connected("bar"))
        self.assertFalse(collection.is_connected("foo"))

        collection.connect("foo")
        self.assertDictEqual(collection.is_connected(), {"bar": False, "foo": True})
        self.assertFalse(collection.is_connected("bar"))
        self.assertTrue(collection.is_connected("foo"))

        self.assertRaises(RuntimeError, collection.connect, "baz")
        self.assertRaises(RuntimeError, collection.disconnect, "baz")

        collection.disconnect()  # Clean up so that the test does not hang

    def test_configuration(self):
        collection = AdapterCollection(
            DummyAdapter("protocol_a", options={"bar": 2, "foo": 3}),
            DummyAdapter("protocol_b", options={"bar": True, "foo": False}),
        )

        self.assertDictEqual(
            collection.configuration(),
            {
                "protocol_a": {"bar": 2, "foo": 3},
                "protocol_b": {"bar": True, "foo": False},
            },
        )

        self.assertDictEqual(
            collection.configuration("protocol_a"),
            {
                "protocol_a": {"bar": 2, "foo": 3},
            },
        )

        self.assertDictEqual(
            collection.configuration("protocol_b"),
            {
                "protocol_b": {"bar": True, "foo": False},
            },
        )

    def test_set_device(self):
        adapter = DummyAdapter(protocol="foo")
        adapter.interface = MagicMock()

        collection = AdapterCollection(adapter)
        collection.set_device("test")

        self.assertEqual(adapter.interface.device, "test")



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
A fluent command builder for lewis.
"""

import re
from functools import partial
from typing import AnyStr

from lewis.adapters.stream import Cmd, regex
from lewis.utils.constants import ACK, ENQ, EOT, ETX, STX

string_arg = partial(str, encoding="utf-8")


class CmdBuilder(object):
    """
    Build a command for the stream adapter.

    Do this by creating this object, adding the values and then building it (this uses a fluent interface).

    For example to read a pressure the ioc might send "pres?" and when that happens this should call get_pres
    command would be:
    >>> CmdBuilder("get_pres").escape("pres?").build()
    This will generate the regex needed by Lewis. The escape is just making sure none of the characters are special
    reg ex characters.
    If you wanted to set a pressure the ioc might send "pres <pressure>" where <pressure> is a floating point number,
    the interface should call set_pres with that number. Now use:
    >>> CmdBuilder("set_pres").escape("pres ").float().build()
    this add float as a regularly expression capture group for your argument. It is equivalent to:
    >>> Cmd("set_pres", "pres ([+-]?\\d+\\.?\\d*)")
    There are various arguments like int and digit. Finally some special characters are included so if your protocol
    uses enquirey character ascii 5 you can match is using
    >>> CmdBuilder("set_pres").escape("pres?").enq().build()
    """

    def __init__(self, target_method, arg_sep="", ignore="", ignore_case=False) -> None:
        """
        Create a builder. Use build to create the final object

        :param target_method: name of the method target to call when the reg ex matches
        :param arg_sep: separators between arguments which are next to each other
        :param ignore: set of characters to ignore between text and arguments
        :param ignore_case: ignore the case when matching command
        """
        self._target_method = target_method
        self._arg_sep = arg_sep
        self._current_sep = ""
        self.argument_mappings = []
        if ignore is None or ignore == "":
            self._ignore = ""
        else:
            self._ignore = "[{0}]*".format(ignore)
        self._reg_ex = self._ignore

        self._ignore_case = ignore_case

    def _add_to_regex(self, regex, is_arg: bool) -> None:
        self._reg_ex += regex + self._ignore
        if not is_arg:
            self._current_sep = ""

    def optional(self, text) -> "CmdBuilder":
        """
        Add some escaped text which does not necessarily need to be there. For commands with optional parameters
        :param text: Text to add
        :return: builder
        """
        self._add_to_regex("(?:" + re.escape(text) + ")?", False)
        return self

    def escape(self, text) -> "CmdBuilder":
        """
        Add some text to the regex which is escaped.

        :param text: text to add
        :return: builder
        """
        self._add_to_regex(re.escape(text), False)
        return self

    def regex(self, new_regex: str) -> "CmdBuilder":
        """
        Add a regex to match but not as an argument.

        :param new_regex: regex to add
        :return: builder
        """
        self._add_to_regex(new_regex, False)
        return self

    def enum(self, *allowed_values: AnyStr) -> "CmdBuilder":
        """
        Matches one of a set of specified strings.

        :param allowed_values: the values this function is allowed to match
        :return: builder
        """
        self._add_to_regex(
            "({})".format("|".join([re.escape(arg) for arg in allowed_values])),
            is_arg=True,
        )
        self.argument_mappings.append(string_arg)
        return self

    def spaces(self, at_least_one: bool = False) -> "CmdBuilder":
        """
        Add a regex for any number of spaces

        :param at_least_one: true there must be at least one space; false there can be any number including zero
        :return: builder

        """
        wildcard = "+" if at_least_one else "*"

        self._add_to_regex(" " + wildcard, False)
        return self

    def arg(self, arg_regex, argument_mapping: partial = string_arg) -> "CmdBuilder":
        """
        Add an argument to the command.

        :param arg_regex: regex for the argument (capture group will be added)
        :param argument_mapping: the type mapping for the argument (default is str)
        :return: builder
        """
        self._add_to_regex(self._current_sep + "(" + arg_regex + ")", True)
        self._current_sep = self._arg_sep
        self.argument_mappings.append(argument_mapping)
        return self

    def string(self, length: None | int = None) -> "CmdBuilder":
        """
        Add an argument which is a string of a given length (if blank string is any length)

        :param length: length of string; None for any length
        :return: builder
        """
        if length is None:
            self.arg(".+")
        else:
            self.arg(".{{{}}}".format(length))
        return self

    def float(self, mapping: type = float, ignore: bool = False) -> "CmdBuilder":
        """
        Add a float argument.

        :param mapping: The type to cast the response to (default: float)
        :param ignore: True to match with a float but ignore the returned value (default: False)
        :return: builder
        """
        regex = r"[+-]?\d+\.?\d*"
        return self.regex(regex) if ignore else self.arg(regex, mapping)

    def digit(self, mapping: type = int, ignore: bool = False) -> "CmdBuilder":
        """
        Add a single digit argument.

        :param mapping: The type to cast the response to (default: int)
        :param ignore: True to match with a digit but ignore the returned value (default: False)
        :return: builder
        """
        return self.regex(r"\d") if ignore else self.arg(r"\d", mapping)

    def char(self, not_chars: None | list[str] | str = None, ignore=False) -> "CmdBuilder":
        """
        Add a single character argument.

        :param not_chars: characters that the character can not be; None for can be anything
        :param ignore: True to match with a char but ignore the returned value (default: False)
        :return: builder
        """
        _regex = r"." if not_chars is None else "[^{}]".format("".join(not_chars))
        return self.regex(_regex) if ignore else self.arg(_regex)

    def int(self, mapping: type = int, ignore: bool = False) -> "CmdBuilder":
        """
        Add an integer argument.

        :param mapping: The type to cast the response to (default: int)
        :param ignore: True to match with a int but ignore the returned value (default: False)
        :return:  builder
        """
        _regex = r"[+-]?\d+"
        return self.regex(_regex) if ignore else self.arg(_regex, mapping)

    def any(self) -> "CmdBuilder":
        """
        Add an argument that matches anything.

        :return: builder
        """
        return self.arg(r".*")

    def any_except(self, char: str) -> "CmdBuilder":
        """
        Adds an argument that matches anything other than a specified character (useful for commands containing
        delimiters)

        :param char: the character not to match
        :return: builder
        """
        return self.arg(r"[^{}]*".format(re.escape(char)))

    def build(self, *args, **kwargs) -> Cmd:
        """
        Builds the CMd object based on the target and regular expression.

        :param args: arguments to pass to Cmd constructor
        :param kwargs: key word arguments to pass to Cmd constructor
        :return: Cmd object
        """
        if self._ignore_case:
            pattern = regex(self._reg_ex)
            pattern.compiled_pattern = re.compile(self._reg_ex.encode(), re.IGNORECASE)
        else:
            pattern = self._reg_ex
        return Cmd(
            self._target_method,
            pattern,
            argument_mappings=self.argument_mappings,
            *args,
            **kwargs,
        )

    def add_ascii_character(self, char_number: int) -> "CmdBuilder":
        """
        Add a single character based on its integer value, e.g. 49 is 'a'.

        :param char_number: character number
        :return: self
        """
        self._add_to_regex(chr(char_number), False)
        return self

    def stx(self) -> "CmdBuilder":
        """
        Add the STX character (0x2) to the string.

        :return: builder
        """
        return self.escape(STX)

    def etx(self) -> "CmdBuilder":
        """
        Add the ETX character (0x3) to the string.

        :return: builder
        """
        return self.escape(ETX)

    def eot(self) -> "CmdBuilder":
        """
        Add the EOT character (0x4) to the string.

        :return: builder
        """
        return self.escape(EOT)

    def enq(self) -> "CmdBuilder":
        """
        Add the ENQ character (0x5) to the string.

        :return: builder
        """
        return self.escape(ENQ)

    def ack(self) -> "CmdBuilder":
        """
        Add the ACK character (0x6) to the string.

        :return: builder
        """
        return self.escape(ACK)

    def eos(self) -> "CmdBuilder":
        """
        Adds the regex end-of-string character to a command.

        :return: builder
        """
        self._reg_ex += "$"
        return self

    def get_multicommands(self, command_separator: AnyStr) -> "CmdBuilder":
        """
        Allows emulator to split multiple commands separated by a defined command separator, e.g. ";".
        Must be accompanied by stream device methods. See Keithley 2700 for examples

        :param command_separator: Character(s) that separate commands
        :return: builder
        """
        self.arg("[^" + re.escape(command_separator) + "]*").escape(command_separator).arg(".*")
        return self



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest
from unittest.mock import Mock, call, patch

import zmq

from lewis.core.control_client import (
    ControlClient,
    ObjectProxy,
    ProtocolException,
    RemoteException,
)


class TestControlClient(unittest.TestCase):
    @patch("zmq.Context")
    def test_zmq_socket_uses_timeout(self, mock_zmq_context):
        timeout = 100
        ControlClient(host="127.0.0.1", port="10002", timeout=timeout)

        mock_zmq_context.assert_has_calls(
            [
                call().setsockopt(zmq.SNDTIMEO, timeout),
                call().setsockopt(zmq.RCVTIMEO, timeout),
            ]
        )

    @patch("uuid.uuid4")
    @patch("lewis.core.control_client.ControlClient._get_zmq_req_socket")
    def test_json_rpc_timeout_raises(self, mock_socket, mock_uuid):
        def zmq_again(self):
            raise zmq.error.Again()

        raising_send = Mock(side_effect=zmq_again)
        mock_uuid.return_value = "2"

        connection = ControlClient(host="127.0.0.1", port="10001")
        connection._socket.send_json = raising_send
        self.assertRaises(ProtocolException, connection.json_rpc, "foo")

        raising_send.assert_called_once_with(
            {"method": "foo", "params": (), "jsonrpc": "2.0", "id": "2"}
        )

    @patch("uuid.uuid4")
    @patch("lewis.core.control_client.ControlClient._get_zmq_req_socket")
    def test_json_rpc(self, mock_socket, mock_uuid):
        mock_uuid.return_value = "2"

        connection = ControlClient(host="127.0.0.1", port="10001")
        connection.json_rpc("foo")

        mock_socket.assert_has_calls(
            [
                call(),
                call().connect("tcp://127.0.0.1:10001"),
                call().send_json({"method": "foo", "params": (), "jsonrpc": "2.0", "id": "2"}),
                call().recv_json(),
            ]
        )

    @patch("lewis.core.control_client.ControlClient._get_zmq_req_socket")
    def test_get_remote_object_works(self, mock_socket):
        client = ControlClient(host="127.0.0.1", port="10001")

        with patch.object(client, "json_rpc") as json_rpc_mock:
            json_rpc_mock.return_value = (
                {
                    "id": 2,
                    "result": {
                        "class": "Test",
                        "methods": ["a:set", "a:get", "setTest"],
                    },
                },
                2,
            )

            obj = client.get_object()

            self.assertTrue(hasattr(type(obj), "a"))
            self.assertTrue(hasattr(obj, "setTest"))

            json_rpc_mock.assert_has_calls([call(":api")])

    @patch("lewis.core.control_client.ControlClient._get_zmq_req_socket")
    def test_get_remote_object_raises_exception(self, mock_socket):
        client = ControlClient(host="127.0.0.1", port="10001")

        with patch.object(client, "json_rpc") as json_rpc_mock:
            json_rpc_mock.return_value = ({"id": 2}, 2)

            self.assertRaises(ProtocolException, client.get_object)

            json_rpc_mock.assert_has_calls([call(":api")])

    @patch("lewis.core.control_client.ControlClient._get_zmq_req_socket")
    def test_get_remote_object_collection(self, mock_socket):
        client = ControlClient(host="127.0.0.1", port="10001")

        returned_object = Mock()
        returned_object.get_objects = Mock(return_value=["obj1", "obj2"])

        with patch.object(client, "get_object") as get_object_mock:
            get_object_mock.side_effect = [
                returned_object,
                "obj1_object",
                "obj2_object",
            ]

            objects = client.get_object_collection()

            self.assertTrue("obj1" in objects)
            self.assertTrue("obj2" in objects)

            returned_object.get_objects.assert_has_calls([call()])
            get_object_mock.assert_has_calls([call(""), call("obj1"), call("obj2")])


class TestObjectProxy(unittest.TestCase):
    def test_init_adds_members(self):
        mock_connection = Mock()
        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["a:get", "a:set", "setTest"])
        self.assertTrue(hasattr(type(obj), "a"))
        self.assertTrue(hasattr(obj, "setTest"))

        mock_connection.assert_not_called()

    def test_member_access_calls_make_request(self):
        obj = type("TestType", (ObjectProxy,), {})(Mock(), ["a:get", "a:set", "setTest"])

        with patch.object(obj, "_make_request") as request_mock:
            obj.a
            obj.a = 4
            obj.setTest()

        request_mock.assert_has_calls([call("a:get"), call("a:set", 4), call("setTest")])

    def test_response_without_id_raises_exception(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = ({"result": "test"}, 2)
        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        self.assertRaises(ProtocolException, obj.setTest)

    def test_response_with_id_mismatch_raises_exception(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = ({"result": "test", "id": 3}, 2)
        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        self.assertRaises(ProtocolException, obj.setTest)

    def test_make_request_with_result(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = ({"result": "test", "id": 2}, 2)
        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        result = obj.setTest()

        self.assertEqual(result, "test")
        mock_connection.json_rpc.assert_has_calls([call("setTest")])

    def test_make_request_with_known_exception(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = (
            {
                "error": {"data": {"type": "AttributeError", "message": "Some message"}},
                "id": 2,
            },
            2,
        )

        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        self.assertRaises(AttributeError, obj.setTest)
        mock_connection.json_rpc.assert_has_calls([call("setTest")])

    def test_make_request_with_unknown_exception(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = (
            {
                "error": {"data": {"type": "NonExistingException", "message": "Some message"}},
                "id": 2,
            },
            2,
        )

        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        self.assertRaises(RemoteException, obj.setTest)
        mock_connection.json_rpc.assert_has_calls([call("setTest")])

    def test_make_request_with_missing_error_data(self):
        mock_connection = Mock(ControlClient)
        mock_connection.json_rpc.return_value = (
            {"error": {"message": "Some message"}},
            2,
        )

        obj = type("TestType", (ObjectProxy,), {})(mock_connection, ["setTest"])

        self.assertRaises(ProtocolException, obj.setTest)
        mock_connection.json_rpc.assert_has_calls([call("setTest")])



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest
from unittest.mock import call, patch

from lewis.core.processor import CanProcess, CanProcessComposite


class TestCanProcess(unittest.TestCase):
    def test_process_calls_doProcess(self):
        processor = CanProcess()

        with patch.object(processor, "doProcess", create=True) as do_process_mock:
            processor.process(1.0)

        do_process_mock.assert_called_once_with(1.0)

    def test_process_calls_doBeforeProcess_only_if_doProcess_is_present(self):
        processor = CanProcess()

        with patch.object(processor, "doBeforeProcess", create=True) as do_before_process_mock:
            processor.process(1.0)

            do_before_process_mock.assert_not_called()

            with patch.object(processor, "doProcess", create=True):
                processor.process(2.0)

            do_before_process_mock.assert_called_once_with(2.0)

    def test_process_calls_doAfterProcess_only_if_doProcess_is_present(self):
        processor = CanProcess()

        with patch.object(processor, "doAfterProcess", create=True) as do_after_process:
            processor.process(1.0)

            do_after_process.assert_not_called()

            with patch.object(processor, "doProcess", create=True):
                processor.process(2.0)

            do_after_process.assert_called_once_with(2.0)

    @patch.object(CanProcess, "process")
    def test_call_invokes_process(self, process_mock):
        processor = CanProcess()

        processor(45.0)

        process_mock.assert_called_once_with(45.0)


class TestCanProcessComposite(unittest.TestCase):
    def test_process_calls_doBeforeProcess_if_present(self):
        composite = CanProcessComposite()

        with patch.object(composite, "doBeforeProcess", create=True) as do_before_process_mock:
            composite.process(3.0)

        do_before_process_mock.assert_called_once_with(3.0)

    def test_addProcessor_if_argument_CanProcess(self):
        composite = CanProcessComposite()

        with patch.object(composite, "_append_processor") as append_processor_mock:
            composite.add_processor(CanProcess())

        self.assertEqual(append_processor_mock.call_count, 1)

    def test_addProcessor_if_argument_not_CanProcess(self):
        composite = CanProcessComposite()

        with patch.object(composite, "_append_processor") as append_processor_mock:
            composite.add_processor(None)

        append_processor_mock.assert_not_called()

    def test_init_from_iterable(self):
        with patch.object(CanProcess, "doProcess", create=True) as mock_process_method:
            devices = (
                CanProcess(),
                CanProcess(),
            )

            composite = CanProcessComposite(devices)
            composite(4.0)

            mock_process_method.assert_has_calls([call(4.0), call(4.0)])



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from lewis.devices.chopper.devices.device import SimulatedChopper
from lewis.devices.chopper.devices.states import DefaultIdleState


class TestSimulatedChopper(unittest.TestCase):
    def test_invalid_state_override_fails(self):
        self.assertRaises(
            RuntimeError,
            SimulatedChopper,
            override_states={"invalid": DefaultIdleState()},
        )

    def test_valid_state_override_does_not_fail(self):
        chopper = SimulatedChopper(override_states={"idle": DefaultIdleState()})

        self.assertIsInstance(chopper, SimulatedChopper)

    def test_invalid_transition_override_fails(self):
        self.assertRaises(
            RuntimeError,
            SimulatedChopper,
            override_transitions={("idle", "phase_locking"): lambda: True},
        )

    def test_valid_transition_override_does_not_fail(self):
        chopper = SimulatedChopper(override_transitions={("idle", "stopping"): lambda: True})

        self.assertIsInstance(chopper, SimulatedChopper)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest
from types import ModuleType
from uuid import uuid4

from lewis.adapters.stream import StreamInterface
from lewis.core.devices import (
    DeviceBase,
    DeviceBuilder,
    DeviceRegistry,
    InterfaceBase,
    is_device,
    is_interface,
)
from lewis.core.exceptions import LewisException
from lewis.devices import Device, StateMachineDevice

from .utils import TestWithPackageStructure, assertRaisesNothing


class TestIsDevice(unittest.TestCase):
    def test_not_a_type_returns_false(self):
        self.assertFalse(is_device(0.0))
        self.assertFalse(is_device(None))

    def test_arbitrary_types_fail(self):
        self.assertFalse(is_device(type(3.0)))
        self.assertFalse(is_device(DeviceBuilder))

    def test_device_bases_are_ignored(self):
        self.assertFalse(is_device(DeviceBase))
        self.assertFalse(is_device(Device))
        self.assertFalse(is_device(StateMachineDevice))

    def test_device_types_work(self):
        class DummyDevice(Device):
            pass

        class DummyStatemachineDevice(StateMachineDevice):
            pass

        self.assertTrue(is_device(DummyDevice))
        self.assertTrue(is_device(DummyStatemachineDevice))


class TestDeviceBuilderSimpleModule(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        class DummyDevice(Device):
            pass

        class DummyInterface(InterfaceBase):
            protocol = "dummy"

        cls.module = ModuleType("simple_dummy_module")
        cls.module.DummyDevice = DummyDevice
        cls.module.DummyAdapter = DummyInterface

    def test_init(self):
        assertRaisesNothing(self, DeviceBuilder, self.module)

        builder = DeviceBuilder(self.module)
        self.assertEqual(builder.name, self.module.__name__)

    def test_defaults(self):
        builder = DeviceBuilder(self.module)

        self.assertIs(builder.default_device_type, self.module.DummyDevice)
        self.assertIs(builder.default_protocol, self.module.DummyAdapter.protocol)

    def test_setups(self):
        builder = DeviceBuilder(self.module)

        setups = builder.setups
        self.assertEqual(len(setups), 1)
        self.assertIn("default", setups)

    def test_protocols(self):
        builder = DeviceBuilder(self.module)

        protocols = builder.protocols
        self.assertEqual(len(protocols), 1)
        self.assertIn("dummy", protocols)

    def test_create_device(self):
        builder = DeviceBuilder(self.module)

        device = builder.create_device()
        self.assertIsInstance(device, self.module.DummyDevice)

        self.assertRaises(LewisException, builder.create_device, "invalid_setup")

    def test_create_interface(self):
        builder = DeviceBuilder(self.module)

        self.assertIsInstance(builder.create_interface(), self.module.DummyAdapter)
        self.assertIsInstance(builder.create_interface("dummy"), self.module.DummyAdapter)

        self.assertRaises(LewisException, builder.create_interface, "invalid_protocol")


class TestDeviceBuilderMultipleDevicesAndProtocols(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        class DummyDevice(Device):
            pass

        class OtherDummyDevice(Device):
            pass

        class DummyInterface(InterfaceBase):
            protocol = "dummy"

        class OtherDummyInterface(InterfaceBase):
            protocol = "other_dummy"

        cls.module = ModuleType("multiple_devices_dummy_module")
        cls.module.DummyDevice = DummyDevice
        cls.module.OtherDummyDevice = OtherDummyDevice
        cls.module.DummyAdapter = DummyInterface
        cls.module.OtherDummyAdapter = OtherDummyInterface

    def test_defaults(self):
        builder = DeviceBuilder(self.module)

        self.assertIs(builder.default_device_type, None)
        self.assertIs(builder.default_protocol, None)

    def test_setups(self):
        builder = DeviceBuilder(self.module)

        setups = builder.setups
        self.assertEqual(len(setups), 1)
        self.assertIn("default", setups)

    def test_protocols(self):
        builder = DeviceBuilder(self.module)

        protocols = builder.protocols
        self.assertEqual(len(protocols), 2)
        self.assertIn("dummy", protocols)
        self.assertIn("other_dummy", protocols)

    def test_create_device(self):
        builder = DeviceBuilder(self.module)

        self.assertRaises(LewisException, builder.create_device)
        self.assertRaises(LewisException, builder.create_device, "default")


class TestDeviceBuilderComplexModule(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        class DummyDevice(Device):
            pass

        class OtherDummyDevice(Device):
            pass

        class DummyInterface(InterfaceBase):
            protocol = "dummy"

        class OtherDummyInterface(InterfaceBase):
            protocol = "other_dummy"

        cls.module = ModuleType("complex_dummy_module")
        cls.module.DummyDevice = DummyDevice
        cls.module.OtherDummyDevice = OtherDummyDevice
        cls.module.interfaces = ModuleType("interfaces")
        cls.module.interfaces.dummy = ModuleType("dummy")
        cls.module.interfaces.dummy.DummyAdapter = DummyInterface
        cls.module.interfaces.other_dummy = ModuleType("other_dummy")
        cls.module.interfaces.other_dummy.OtherDummyAdapter = OtherDummyInterface
        cls.module.setups = ModuleType("setups")
        cls.module.setups.default = ModuleType("default")
        cls.module.setups.default.device_type = DummyDevice
        cls.module.setups.other = ModuleType("other")
        cls.module.setups.other.device_type = OtherDummyDevice

    def test_defaults(self):
        builder = DeviceBuilder(self.module)

        self.assertIs(builder.default_device_type, None)
        self.assertIs(builder.default_protocol, None)

    def test_setups(self):
        builder = DeviceBuilder(self.module)

        setups = builder.setups
        self.assertEqual(len(setups), 2)
        self.assertIn("default", setups)
        self.assertIn("other", setups)

    def test_create_device(self):
        builder = DeviceBuilder(self.module)

        self.assertIsInstance(builder.create_device(), self.module.DummyDevice)
        self.assertIsInstance(builder.create_device("default"), self.module.DummyDevice)
        self.assertIsInstance(builder.create_device("other"), self.module.OtherDummyDevice)


class TestDeviceBuilderWithDuplicateProtocols(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        class DummyInterface(InterfaceBase):
            protocol = "dummy"

        class DummyInterfaceTwo(InterfaceBase):
            protocol = "dummy"

        cls.module = ModuleType("simple_dummy_module")
        cls.module.DummyAdapter = DummyInterface
        cls.module.DummyAdapterTwo = DummyInterfaceTwo

    def test_init_fails(self):
        self.assertRaises(RuntimeError, DeviceBuilder, self.module)


class TestDeviceRegistry(TestWithPackageStructure):
    def test_init(self):
        assertRaisesNothing(self, DeviceRegistry, self._tmp_package_name)
        self.assertRaises(LewisException, DeviceRegistry, str(uuid4()))

    def test_devices(self):
        registry = DeviceRegistry(self._tmp_package_name)

        devices = registry.devices
        self.assertEqual(len(devices), 2)
        self.assertIn("some_file", devices)
        self.assertIn("some_dir", devices)

    def test_device_builder(self):
        registry = DeviceRegistry(self._tmp_package_name)

        builder = registry.device_builder("some_file")
        self.assertEqual(builder.name, "some_file")
        self.assertRaises(LewisException, registry.device_builder, "invalid_device")


class TestIsInterface(unittest.TestCase):
    def test_not_a_type_returns_false(self):
        self.assertFalse(is_interface(0.0))
        self.assertFalse(is_interface(None))

    def test_arbitrary_types_fail(self):
        self.assertFalse(is_interface(type(3.0)))
        self.assertFalse(is_interface(property))

    def test_interface_base_is_ignored(self):
        self.assertFalse(is_interface(InterfaceBase))
        self.assertFalse(is_interface(StreamInterface))

    def test_interface_types_work(self):
        class DummyInterface(InterfaceBase):
            pass

        self.assertTrue(is_interface(DummyInterface))



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from collections import OrderedDict

from lewis.core.utils import check_limits
from lewis.devices import StateMachineDevice

from . import states


class SimulatedJulabo(StateMachineDevice):
    internal_p = 0.1  # The proportional
    internal_i = 3  # The integral
    internal_d = 0  # The derivative
    external_p = 0.1  # The proportional
    external_i = 3  # The integral
    external_d = 0  # The derivative
    temperature_low_limit = 0.0  # Usually set in the hardware
    temperature_high_limit = 100.0  # Usually set in the hardware
    set_point_temperature = 24.0  # Set point starts equal to the current temperature
    heating_power = 5.0  # The heating power
    version = "JULABO FP50_MH Simulator, ISIS"
    status = "Hello from the simulated Julabo"
    is_circulating = 0  # 0 for off, 1 for on
    temperature = 24.0  # Current temperature in C
    external_temperature = 26.0  # External temperature in C
    circulate_commanded = False
    temperature_ramp_rate = 5.0  # Guessed value in C/min

    def _initialize_data(self) -> None:
        """
        This method is called once on construction. After that, it may be
        manually called again to reset the device to its default state.

        After the first call during construction, the class is frozen.

        This means that attempting to define a new member variable will
        raise an exception. This is to prevent typos from inadvertently
        and silently adding new members instead of accessing existing ones.
        """
        pass

    def _get_state_handlers(self):
        return {
            "circulate": states.DefaultCirculatingState(),
            "not_circulate": states.DefaultNotCirculatingState(),
        }

    def _get_initial_state(self) -> str:
        return "not_circulate"

    def _get_transition_handlers(self):
        return OrderedDict(
            [
                (("not_circulate", "circulate"), lambda: self.circulate_commanded),
                (("circulate", "not_circulate"), lambda: not self.circulate_commanded),
            ]
        )

    def set_set_point(self, param) -> str:
        """
        Sets the target temperature.

        :param param: The new temperature in C. Must be positive.
        :return: Empty string.
        """
        if self.temperature_low_limit <= param <= self.temperature_high_limit:
            self.set_point_temperature = param
        return ""

    def set_circulating(self, param) -> str:
        """
        Sets whether to circulate - in effect whether the heater is on.

        :param param: The mode to set, must be 0 or 1.
        :return: Empty string.
        """
        if param == 0:
            self.is_circulating = param
            self.circulate_commanded = False
        elif param == 1:
            self.is_circulating = param
            self.circulate_commanded = True
        return ""

    @check_limits(0.1, 99.9)
    def set_internal_p(self, param) -> str:
        """
        Sets the internal proportional.
        Xp in Julabo speak.

        :param param: The value to set, must be between 0.1 and 99.9
        :return: Empty string.
        """
        self.internal_p = param
        return ""

    @check_limits(3, 9999)
    def set_internal_i(self, param) -> str:
        """
        Sets the internal integral.
        Tn in Julabo speak.

        :param param: The value to set, must be an integer between 3 and 9999
        :return: Empty string.
        """
        self.internal_i = param
        return ""

    @check_limits(0, 999)
    def set_internal_d(self, param) -> str:
        """
        Sets the internal derivative.
        Tv in Julabo speak.

        :param param: The value to set, must be an integer between 0 and 999
        :return: Empty string.
        """
        self.internal_d = param
        return ""

    @check_limits(0.1, 99.9)
    def set_external_p(self, param) -> str:
        """
        Sets the external proportional.
        Xp in Julabo speak.

        :param param: The value to set, must be between 0.1 and 99.9
        :return: Empty string.
        """
        self.external_p = param
        return ""

    @check_limits(3, 9999)
    def set_external_i(self, param) -> str:
        """
        Sets the external integral.
        Tn in Julabo speak.

        :param param: The value to set, must be an integer between 3 and 9999
        :return: Empty string.
        """
        self.external_i = param
        return ""

    @check_limits(0, 999)
    def set_external_d(self, param) -> str:
        """
        Sets the external derivative.
        Tv in Julabo speak.

        :param param: The value to set, must be an integer between 0 and 999
        :return: Empty string.
        """
        self.external_d = param
        return ""



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from lewis.core.approaches import linear


class TestApproachLinear(unittest.TestCase):
    def test_target_equals_current_does_not_change(self):
        pos = 34.0
        self.assertEqual(linear(pos, pos, 15.0, 0.5), pos)

    def test_speed_zero_does_not_change_value(self):
        pos = 34.0
        target = 23.0
        self.assertEqual(linear(pos, target, 0.0, 0.5), pos)

    def test_dt_zero_does_not_change_value(self):
        pos = 34.0
        target = 23.0
        self.assertEqual(linear(pos, target, 1.0, 0.0), pos)

    def test_target_less_than_pos_works(self):
        pos = 34.0
        target = 23.0
        self.assertEqual(linear(pos, target, 2.0, 0.5), 33.0)

    def test_target_greater_than_pos_works(self):
        pos = 34.0
        target = 43.0
        self.assertEqual(linear(pos, target, 2.0, 0.5), 35.0)

    def test_target_negative_speed_inverts_behavior(self):
        pos = 34.0
        self.assertEqual(linear(pos, 23, -2.0, 0.5), 35.0)
        self.assertEqual(linear(pos, 43, -2.0, 0.5), 33.0)

    def test_no_overshoot(self):
        pos = 34.0
        target = 33.0
        self.assertEqual(linear(pos, target, 10.0, 100.0), target)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains base classes for devices. Inherit from :class:`Device` for simple devices
or from :class:`StateMachineDevice` for devices that are more complex and can be described
using a state machine.
"""

from logging import Logger
from typing import Callable

from lewis.core.devices import DeviceBase
from lewis.core.processor import CanProcess, CanProcessComposite
from lewis.core.statemachine import State, StateMachine
from lewis.core.utils import dict_strict_update


class Device(DeviceBase, CanProcess):
    """
    This class exists mainly for consistency. It is meant to implement very simple devices that
    do not require a state machine for their simulation. For such devices, all that is required
    is subclassing from `Device` and possibly implementing `doProcess`, but this is optional.

    StateMachineDevice offers more functionality and is more likely to be useful for implementing
    simulations of real devices.
    """

    def __init__(self) -> None:
        super(Device, self).__init__()


class StateMachineDevice(DeviceBase, CanProcessComposite):
    """
    This class is intended to be sub-classed to implement devices using a finite state machine
    internally.

    Implementing such a device is straightforward, there are three methods
    that *must* be overridden:

        - :meth:`_get_state_handlers`
        - :meth:`_get_initial_state`
        - :meth:`_get_transition_handlers`

    The first method is supposed to return a dictionary with state handlers for each state
    of the state machine, the second method must return the name of the initial state.
    The third method must return a dict-like object (often an OrderedDict from collections)
    that defines the conditions for transitions between the states of the state machine.

    They are implemented as methods and not as plain class member variables, because often
    they use the `self`-variable, which does not exist at the class level.

    From these three methods, a :class:`~lewis.core.statemachine.StateMachine`-instance is
    constructed, it's available as the device's ``_csm``-member. CSM is short for
    "cycle-based state machine".

    Most device implementation will also want to override this method:

        - :meth:`_initialize_data`

    This method should initialise device state variables (such as temperature, speed, etc.).
    Having this in a separate method from ``__init__`` has the advantage that it can be used
    to reset those variables at a later stage, without having to write the same code again.

    Following this scheme, inheriting from StateMachineDevice also provides the possibility
    for users of the class to override the states, the transitions, the initial state and
    even the data. For states, transitions and data, dicts need to be passed to the
    constructor, for the initial state that should be a string.

    All these overrides can be used to define device setups to describe certain scenarios
    more easily.

    :param override_states: Dict with one entry per state. Only states defined in the state
                            machine are allowed.
    :param override_transitions: Dict with (state, state) tuples as keys and
                                 callables as values.
    :param override_initial_state: The initial state.
    :param override_initial_data: A dict that contains data members
                                  that should be overwritten on construction.
    """

    def __init__(
        self,
        override_states: dict[str, State] | None = None,
        override_transitions: dict[tuple[State, State], Callable[[], bool]] | None = None,
        override_initial_state: State | None = None,
        override_initial_data: dict[str, float] | None = None,
    ) -> None:
        super(StateMachineDevice, self).__init__()

        self.log: Logger
        self.log.info("Creating device, setting up state machine")

        self._initialize_data()
        self._override_data(override_initial_data)

        state_handlers = self._get_final_state_handlers(override_states)
        initial = override_initial_state or self._get_initial_state()

        if initial not in state_handlers:
            raise RuntimeError("Initial state '{}' is not a valid state.".format(initial))

        self._csm = StateMachine(
            {
                "initial": initial,
                "states": state_handlers,
                "transitions": self._get_final_transition_handlers(override_transitions),
            },
            context=self,
        )

        self.add_processor(self._csm)

    def _get_state_handlers(self) -> dict[str, State]:
        """
        Implement this method to return a dict-like object with state handlers
        (see :class:`~lewis.core.statemachine.State`) for each state of the state machine.
        The default implementation raises a ``NotImplementedError``.

        :return: A dict-like object containing named state handlers.
        """
        raise NotImplementedError(
            "_get_state_handlers must be implemented in a StateMachineDevice."
        )

    def _get_initial_state(self) -> str:
        """
        Implement this method to return the initial state of the internal state machine.
        The default implementation raises a ``NotImplementedError``.

        :return: The initial state of the state machine.
        """
        raise NotImplementedError("_get_initial_state must be implemented in a StateMachineDevice.")

    def _get_transition_handlers(self) -> dict[tuple[str, str], Callable[[], bool]]:
        """
        Implement this method to return transition handlers for the internal state machine.
        The keys should be (state, state)-tuples and the values functions that return true
        if the transition should be triggered. The default implementation raises a
        ``NotImplementedError``.

        :return: A dict-like object containing transition handlers.
        """
        raise NotImplementedError(
            "_get_transition_handlers must be implemented in a StateMachineDevice."
        )

    def _initialize_data(self) -> None:
        """
        Implement this method to initialize data members of the device, such as temperature,
        speed and others. It gets called first in the __init__-method. The default implementation
        does nothing.
        """
        pass

    def _get_final_state_handlers(self, overrides: dict[str, State] | None) -> dict[str, State]:
        states = self._get_state_handlers()

        if overrides is not None:
            dict_strict_update(states, overrides)

        return states

    def _get_final_transition_handlers(
        self, overrides: dict[tuple[State, State], Callable[[], bool]] | None
    ) -> dict[tuple[State, State], Callable[[], bool]]:
        transitions = self._get_transition_handlers()

        if overrides is not None:
            dict_strict_update(transitions, overrides)

        return transitions

    def _override_data(self, overrides: dict[str, float] | None) -> None:
        """
        This method overrides data members of the class, but does not allow for adding new members.

        :param overrides: Dict with data overrides.
        """
        if overrides is not None:
            for name, val in overrides.items():
                self.log.debug("Trying to override initial data (%s=%s)", name, val)
                if name not in dir(self):
                    raise AttributeError(
                        "Can not override non-existing attribute'{}' of class '{}'.".format(
                            name, type(self).__name__
                        )
                    )

                setattr(self, name, val)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from parameterized import parameterized

from lewis.adapters.stream import Func


class TestFunc(unittest.TestCase):
    """Unit tests for lewis.adapters.stream.Func"""

    def setUp(self):
        self.target_member = lambda x: 7
        self.write_pattern = "([0-8]{1})"
        self.argument_mapping = [int]
        self.incorrect_argument_mapping = int
        self.return_mapping = int
        self.process_request_value = b"7"

    @parameterized.expand(
        [
            ("valid_regex", lambda: 0, "[0-9]{1,2}", None),
            ("valid_regex_and_argument_mapping", lambda x: 5, "([0-8]{1})", [int]),
        ]
    )
    def test_argument_variations_of_valid_Func_without_return_mapping_is_instance_of_Func(
        self, _, target_member, write_pattern, argument_mapping
    ):
        self.assertIsInstance(Func(target_member, write_pattern, argument_mapping), Func)

    @parameterized.expand(
        [
            ("invalid_function", "invalid_func", "[0-9]{1,2}", int, None),
            ("invalid_regex", lambda x: 5, "[0-9]*{1}", int, None),
            ("invalid_argument", lambda: 0, "[0-9]{1}", [1], None),
            ("invalid_argument_and_return_mappings", lambda: 0, "[0-9]{1}", [1], 1),
        ]
    )
    def test_argument_variations_of_Func_usage(
        self, _, target_member, write_pattern, argument_mapping, return_mapping
    ):
        with self.assertRaises(RuntimeError):
            Func(target_member, write_pattern, argument_mapping, return_mapping)

    def test_argument_invalid_return_mapping_type_returns_TypeError(self):
        invalid_return_mapping = 1
        with self.assertRaises(TypeError):
            Func(
                self.target_member,
                self.write_pattern,
                self.incorrect_argument_mapping,
                invalid_return_mapping,
            )

    def test_func_returns_correct_value_for_return_argument_mapping(self):
        self.assertEqual(
            Func(
                self.target_member,
                self.write_pattern,
                self.argument_mapping,
                self.return_mapping,
            ).process_request(self.process_request_value),
            7,
        )

    def test_process_request_is_int(self):
        self.assertIsInstance(
            Func(
                self.target_member,
                self.write_pattern,
                self.argument_mapping,
                self.return_mapping,
            ).process_request(self.process_request_value),
            int,
        )

    def test_func_fails_for_incorrect_return_argument_mappings(self):
        with self.assertRaises(ValueError):
            Func(
                lambda x: "string",
                self.write_pattern,
                self.argument_mapping,
                self.return_mapping,
            ).process_request(self.process_request_value)

    def test_process_request(self):
        self.assertTrue(
            Func(
                self.target_member,
                self.write_pattern,
                self.argument_mapping,
                self.return_mapping,
            ).can_process(self.process_request_value)
        )

    def test_incorrect_can_process_request_fails(self):
        self.assertFalse(
            Func(
                self.target_member,
                self.write_pattern,
                self.argument_mapping,
                self.return_mapping,
            ).can_process(b"9")
        )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import itertools
import unittest
from unittest.mock import MagicMock, Mock, call, patch

from lewis.devices import StateMachineDevice

from .utils import assertRaisesNothing


class MockStateMachineDevice(StateMachineDevice):
    _get_state_handlers = Mock(return_value={"init": {}, "test": {}})
    _get_initial_state = Mock(return_value="init")
    _get_transition_handlers = Mock(return_value={})
    _initialize_data = Mock()

    existing_member = 1.0


class TestStateMachineDevice(unittest.TestCase):
    def test_not_implemented_errors(self):
        # Construction of the base class should not be possible
        self.assertRaises(NotImplementedError, StateMachineDevice)

        mandatory_methods = {
            "_get_state_handlers": Mock(return_value={"test": MagicMock()}),
            "_get_initial_state": Mock(return_value="test"),
            "_get_transition_handlers": Mock(
                return_value={("test", "test"): Mock(return_value=True)}
            ),
        }

        # If any of the mandatory methods is missing, a NotImplementedError must be raised
        for methods in itertools.combinations(mandatory_methods.items(), 2):
            with patch.multiple("lewis.devices.StateMachineDevice", **dict(methods)):
                self.assertRaises(NotImplementedError, StateMachineDevice)

        # If all are implemented, no exception should be raised
        with patch.multiple("lewis.devices.StateMachineDevice", **mandatory_methods):
            assertRaisesNothing(self, StateMachineDevice)

    def test_init_calls_appropriate_methods(self):
        smd = MockStateMachineDevice()

        smd._get_state_handlers.assert_has_calls([call()])
        smd._get_initial_state.assert_has_calls([call()])
        smd._get_transition_handlers.assert_has_calls([call()])
        smd._initialize_data.assert_has_calls([call()])

    def test_invalid_initial_override_fails(self):
        assertRaisesNothing(self, MockStateMachineDevice)
        assertRaisesNothing(self, MockStateMachineDevice, override_initial_state="init")
        assertRaisesNothing(self, MockStateMachineDevice, override_initial_state="test")

        self.assertRaises(RuntimeError, MockStateMachineDevice, override_initial_state="invalid")

    def test_overriding_undefined_data_fails(self):
        assertRaisesNothing(
            self, MockStateMachineDevice, override_initial_data={"existing_member": 2.0}
        )

        smd = MockStateMachineDevice(override_initial_data={"existing_member": 2.0})
        self.assertEqual(smd.existing_member, 2.0)

        self.assertRaises(
            AttributeError,
            MockStateMachineDevice,
            override_initial_data={"nonexisting_member": 1.0},
        )



import unittest
from unittest.mock import MagicMock, patch

from parameterized import parameterized

from lewis.adapters.stream import StreamHandler


@patch("asynchat.async_chat")
class TestStreamHandler(unittest.TestCase):
    def setUp(self):
        """Create a mock for the async_chat class"""
        self.target = MagicMock()
        self.stream_server = MagicMock()
        self.socket = MagicMock()
        self.handler = StreamHandler(self.socket, self.target, self.stream_server)

    @parameterized.expand(
        [
            (b"\n", "test", b"test\n"),
            (b"\n", b"test", b"test\n"),
            ("\n", "test", b"test\n"),
            ("\n", "test", b"test\n"),
        ]
    )
    @patch("asynchat.async_chat.push")
    def test_terminator_and_replies_of_different_types_can_be_concatenated(
        self, terminator, message, expected, async_push, _
    ):
        self.target.out_terminator = terminator
        self.handler.unsolicited_reply(message)

        self.assertEqual(expected, async_push.call_args[0][0])
